# generated by onnxconverter_common.onnx2py

"""
Run this script to recreate the original onnx model.
Example usage:
python matmulnbits_add_gen.py out_model_path.onnx
"""

import os
import sys

import numpy as np
import onnx
from onnx import TensorProto, helper, numpy_helper

DATA_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), "matmulnbits_add_gen")


def clear_field(proto, field):
    proto.ClearField(field)
    return proto


def order_repeated_field(repeated_proto, key_name, order):
    order = list(order)
    repeated_proto.sort(key=lambda x: order.index(getattr(x, key_name)))


def make_node(op_type, inputs, outputs, name=None, doc_string=None, domain=None, **kwargs):
    node = helper.make_node(op_type, inputs, outputs, name, doc_string, domain, **kwargs)
    if doc_string == "":
        node.doc_string = ""
    order_repeated_field(node.attribute, "name", kwargs.keys())
    return node


def make_graph(*args, doc_string=None, **kwargs):
    graph = helper.make_graph(*args, doc_string=doc_string, **kwargs)
    if doc_string == "":
        graph.doc_string = ""
    return graph


model = helper.make_model(
    opset_imports=[
        clear_field(helper.make_operatorsetid("", 21), "domain"),
        helper.make_operatorsetid("com.microsoft", 1),
    ],
    ir_version=10,
    graph=make_graph(
        name="MatMul_Add",
        inputs=[
            helper.make_tensor_value_info("A", TensorProto.FLOAT, shape=["M", 2]),
            helper.make_tensor_value_info("bias", TensorProto.FLOAT, shape=[3]),
        ],
        outputs=[helper.make_tensor_value_info("C", TensorProto.FLOAT, shape=["M", 3])],
        initializer=[
            numpy_helper.from_array(
                np.load(os.path.join(DATA_DIR, "const0_B_Q4.npy")).astype("uint8").reshape([3, 1, 16]), name="B_Q4"
            ),
            numpy_helper.from_array(np.array([-0.125, -0.125, -0.125], dtype="float32"), name="B_scales"),
        ],
        nodes=[
            make_node(
                "MatMulNBits",
                inputs=["A", "B_Q4", "B_scales"],
                outputs=["matmul_out"],
                name="matmul_Q4",
                domain="com.microsoft",
                K=2,
                N=3,
                bits=4,
                block_size=32,
            ),
            make_node("Add", inputs=["matmul_out", "bias"], outputs=["C"], name="add"),
        ],
    ),
)

if __name__ == "__main__" and len(sys.argv) == 2:
    _, out_path = sys.argv
    onnx.save(model, out_path)



#!/usr/bin/env python3

# args: <submodule path>
# output: <path> <url> <commit>

import subprocess
import sys

assert len(sys.argv) == 2

path = sys.argv[1]

proc = subprocess.run(
    ["git", "config", "--get", "remote.origin.url"],
    check=True,
    cwd=path,
    stdout=subprocess.PIPE,
    text=True,
)

url = proc.stdout.strip()

proc = subprocess.run(["git", "rev-parse", "HEAD"], check=True, cwd=path, stdout=subprocess.PIPE, text=True)

commit = proc.stdout.strip()

print(f"{path} {url} {commit}")



#!/usr/bin/env python
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
from __future__ import annotations

import itertools
import os
import tempfile
import unittest

import numpy as np
import onnx
from onnx import TensorProto, helper
from op_test_utils import (
    TestDataFeeds,
    check_model_correctness,
    check_op_type_count,
    check_qtype_by_node_type,
    get_tensor_consumers_and_producers,
)

from onnxruntime.quantization import QuantFormat, QuantType, quantize_dynamic, quantize_static


class TestOpQuatizerPad(unittest.TestCase):
    def input_feeds(self, n, name2shape):
        input_data_list = []
        for _i in range(n):
            inputs = {}
            for name, shape in name2shape.items():
                inputs.update({name: np.random.randint(-1, 2, shape).astype(np.float32)})
            input_data_list.extend([inputs])
        dr = TestDataFeeds(input_data_list)
        return dr

    def construct_model_pad(
        self,
        output_model_path,
        pad_mode,
        pad_input_shape,
        pad_dims,
        constant_value=None,
    ):
        #    (input)
        #      |
        #     Pad
        #      |
        #    (output)
        rank = len(pad_input_shape)
        self.assertEqual(rank * 2, len(pad_dims))

        input_tensor = helper.make_tensor_value_info("input", TensorProto.FLOAT, pad_input_shape)
        pad_dims_initializer = helper.make_tensor("pad_dims", TensorProto.INT64, [2 * rank], pad_dims)
        output_shape = [sum(e) for e in list(zip(pad_input_shape, pad_dims[:rank], pad_dims[rank:], strict=False))]
        output_tensor = helper.make_tensor_value_info("output", TensorProto.FLOAT, output_shape)

        inputs = ["input", "pad_dims"]
        initializers = [pad_dims_initializer]
        if (constant_value is not None) and (pad_mode is None or pad_mode == "constant"):
            constant_value_tensor = helper.make_tensor("padding_value", TensorProto.FLOAT, [], [constant_value])
            inputs.extend(["padding_value"])
            initializers.extend([constant_value_tensor])
        kwargs = {"mode": pad_mode} if pad_mode is not None else {}
        pad_node = helper.make_node("Pad", inputs, ["output"], name="PadNode", **kwargs)

        graph = helper.make_graph(
            [pad_node],
            "TestOpQuantizerPad_test_model",
            [input_tensor],
            [output_tensor],
            initializer=initializers,
        )
        model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", 13)])
        model.ir_version = 7  # use stable onnx ir version

        onnx.save(model, output_model_path)

    def construct_model_conv_pad(
        self,
        output_model_path,
        conv_input_shape,
        conv_weight_shape,
        pad_input_shape,
        pad_mode,
        pad_dims,
        constant_value=None,
    ):
        #      (input)
        #          \
        #         Conv
        #        /    \
        #   Identity   Pad
        #    /            \
        # (identity_out)  (output)
        rank = len(pad_input_shape)
        self.assertEqual(rank * 2, len(pad_dims))

        input_tensor = helper.make_tensor_value_info("input", TensorProto.FLOAT, conv_input_shape)

        conv_weight_arr = np.random.randint(-1, 2, conv_weight_shape).astype(np.float32)
        conv_weight_initializer = onnx.numpy_helper.from_array(conv_weight_arr, name="conv1_weight")
        conv_node = onnx.helper.make_node("Conv", ["input", "conv1_weight"], ["conv_output"], name="conv_node")

        identity_out = helper.make_tensor_value_info("identity_out", TensorProto.FLOAT, pad_input_shape)
        identity_node = helper.make_node("Identity", ["conv_output"], ["identity_out"], name="IdentityNode")

        pad_dims_initializer = helper.make_tensor("pad_dims", TensorProto.INT64, [2 * rank], pad_dims)
        output_shape = [sum(e) for e in list(zip(pad_input_shape, pad_dims[:rank], pad_dims[rank:], strict=False))]
        output_tensor = helper.make_tensor_value_info("output", TensorProto.FLOAT, output_shape)
        pad_inputs = ["conv_output", "pad_dims"]
        initializers = [conv_weight_initializer, pad_dims_initializer]
        if (constant_value is not None) and (pad_mode is None or pad_mode == "constant"):
            constant_value_tensor = helper.make_tensor("padding_value", TensorProto.FLOAT, [], [constant_value])
            pad_inputs.extend(["padding_value"])
            initializers.extend([constant_value_tensor])
        kwargs = {"mode": pad_mode} if pad_mode is not None else {}
        pad_node = helper.make_node("Pad", pad_inputs, ["output"], name="pad_node", **kwargs)

        graph = helper.make_graph(
            [conv_node, identity_node, pad_node],
            "TestOpQuantizerPad_test_model",
            [input_tensor],
            [identity_out, output_tensor],
            initializer=initializers,
        )
        model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", 13)])
        model.ir_version = 7  # use stable onnx ir version
        onnx.save(model, output_model_path)

    def quantize_model(
        self,
        model_fp32_path,
        model_i8_path,
        data_reader=None,
        activation_type=QuantType.QUInt8,
        weight_type=QuantType.QUInt8,
        extra_options={},  # noqa: B006
    ):
        if data_reader is not None:
            quantize_static(
                model_fp32_path,
                model_i8_path,
                data_reader,
                reduce_range=True,
                quant_format=QuantFormat.QOperator,
                activation_type=activation_type,
                weight_type=weight_type,
                extra_options=extra_options,
            )
        else:
            quantize_dynamic(
                model_fp32_path,
                model_i8_path,
                reduce_range=True,
                weight_type=weight_type,
                extra_options=extra_options,
            )

    def verify_should_not_trigger(self, quantize_mode="static"):
        np.random.seed(108)
        model_fp32_path = f"qop_pad_notrigger_fp32_{quantize_mode}.onnx"
        model_i8_path = f"qop_pad_notrigger_i8_{quantize_mode}.onnx"
        data_reader = self.input_feeds(1, {"input": [1, 16, 31, 31]})
        self.construct_model_pad(model_fp32_path, "constant", [1, 16, 31, 31], [0, 0, 1, 2, 0, 0, 3, 4])
        self.quantize_model(
            model_fp32_path,
            model_i8_path,
            None if quantize_mode != "static" else data_reader,
        )
        data_reader.rewind()
        # DequantizeLinear=0 pad node is not been quantized as input is not quantized.
        check_op_type_count(
            self,
            model_i8_path,
            DynamicQuantizeLinear=0,
            QuantizeLinear=0,
            DequantizeLinear=0,
        )
        check_model_correctness(self, model_fp32_path, model_i8_path, data_reader.get_next())

    def test_static_quantize_no_trigger(self):
        self.verify_should_not_trigger(quantize_mode="static")

    def test_dynamic_quantize_no_trigger(self):
        self.verify_should_not_trigger(quantize_mode="dynamic")

    def verify_quantize_with_pad_mode(
        self,
        pad_mode,
        constant_value=None,
        quantize_mode="static",
        rtol=0.01,
        atol=0.05,
        activation_type=QuantType.QUInt8,
        weight_type=QuantType.QUInt8,
        extra_options=None,
    ):
        if extra_options is None:
            extra_options = {}
        np.random.seed(108)
        tag_pad_mode = pad_mode if pad_mode is not None else "none"
        tag_constant_value = "" if constant_value is None else "_value"
        model_fp32_path = f"qop_pad_{quantize_mode}_fp32_{tag_pad_mode}{tag_constant_value}.onnx"
        edge_case = "dual_feed" in extra_options and extra_options["dual_feed"] and constant_value is not None
        if edge_case:
            data_reader = self.input_feeds(1, {"input": [1, 8, 33, 33], "padding_value": [1]})
            self.construct_edge_case_model(
                model_fp32_path,
                [1, 8, 33, 33],
                [16, 8, 3, 3],
                [1, 16, 31, 31],
                pad_mode,
                [0, 0, 1, 2, 0, 0, 3, 4],
                constant_value=constant_value,
            )
        else:
            data_reader = self.input_feeds(1, {"input": [1, 8, 33, 33]})
            self.construct_model_conv_pad(
                model_fp32_path,
                [1, 8, 33, 33],
                [16, 8, 3, 3],
                [1, 16, 31, 31],
                pad_mode,
                [0, 0, 1, 2, 0, 0, 3, 4],
                constant_value=constant_value,
            )

        activation_proto_qtype = TensorProto.UINT8 if activation_type == QuantType.QUInt8 else TensorProto.INT8
        activation_type_str = "u8" if (activation_type == QuantType.QUInt8) else "s8"
        weight_type_str = "u8" if (weight_type == QuantType.QUInt8) else "s8"
        model_i8_path = (
            f"qop_pad_{quantize_mode}_i8_{tag_pad_mode}{tag_constant_value}_{activation_type_str}{weight_type_str}.onnx"
        )
        data_reader.rewind()
        self.quantize_model(
            model_fp32_path,
            model_i8_path,
            None if quantize_mode != "static" else data_reader,
            activation_type=activation_type,
            weight_type=weight_type,
            extra_options=extra_options,
        )
        # DequantizeLinear=2 means there are one DequantizeLinear Node aftr both conv and pad,
        # which means pad node is running in quantized semantic.
        # In dynamic quantize mode, pad operator in fact not quantized as input is fp32.
        if quantize_mode != "static":
            kwargs = {"DynamicQuantizeLinear": 1} if activation_type == QuantType.QUInt8 else {"QuantizeLinear": 1}
        else:
            # edge case will have 2 graph inputs
            kwargs = {"DequantizeLinear": 2, "QuantizeLinear": 2 if edge_case else 1}
        check_op_type_count(self, model_i8_path, **kwargs)
        # check node input/output type if such node exists in the graph
        qnode_io_qtypes = {
            "QuantizeLinear": [
                ["i", 2, activation_proto_qtype],
                ["o", 0, activation_proto_qtype],
            ]
        }
        qnode_io_qtypes.update({"DequantizeLinear": [["i", 2, activation_proto_qtype]]})
        qnode_io_qtypes.update({"ConvInteger": [["i", 2, activation_proto_qtype]]})
        check_qtype_by_node_type(self, model_i8_path, qnode_io_qtypes)
        data_reader.rewind()
        check_model_correctness(
            self,
            model_fp32_path,
            model_i8_path,
            data_reader.get_next(),
            rtol=rtol,
            atol=atol,
        )

    def test_static_mode_edge(self):
        self.verify_quantize_with_pad_mode("edge", constant_value=None)

    def test_static_mode_reflect(self):
        self.verify_quantize_with_pad_mode("reflect", constant_value=None)

    def test_static_mode_constant_default(self):
        self.verify_quantize_with_pad_mode("constant", constant_value=None)

    def test_static_mode_constant_value(self):
        self.verify_quantize_with_pad_mode("constant", constant_value=3.75)

    def test_static_mode_edge_s8s8(self):
        self.verify_quantize_with_pad_mode(
            "edge",
            constant_value=None,
            rtol=0.1,
            atol=0.1,
            activation_type=QuantType.QInt8,
            weight_type=QuantType.QInt8,
            extra_options={"ActivationSymmetric": True},
        )

    def test_static_mode_reflect_s8s8(self):
        self.verify_quantize_with_pad_mode(
            "reflect",
            constant_value=None,
            rtol=0.1,
            atol=0.1,
            activation_type=QuantType.QInt8,
            weight_type=QuantType.QInt8,
            extra_options={"ActivationSymmetric": True},
        )

    def test_static_mode_constant_default_s8s8(self):
        self.verify_quantize_with_pad_mode(
            "constant",
            constant_value=None,
            rtol=0.1,
            atol=0.1,
            activation_type=QuantType.QInt8,
            weight_type=QuantType.QInt8,
            extra_options={"ActivationSymmetric": True},
        )

    def test_static_mode_constant_value_s8s8(self):
        self.verify_quantize_with_pad_mode(
            "constant",
            constant_value=3.75,
            rtol=0.1,
            atol=0.1,
            activation_type=QuantType.QInt8,
            weight_type=QuantType.QInt8,
            extra_options={"ActivationSymmetric": True},
        )

    def test_dynamic_mode_edge(self):
        self.verify_quantize_with_pad_mode("edge", constant_value=None, quantize_mode="dynamic")

    def test_dynamic_mode_reflect(self):
        self.verify_quantize_with_pad_mode("reflect", constant_value=None, quantize_mode="dynamic")

    def test_dynamic_mode_constant_default(self):
        self.verify_quantize_with_pad_mode("constant", constant_value=None, quantize_mode="dynamic")

    def test_dynamic_mode_constant_value(self):
        self.verify_quantize_with_pad_mode("constant", constant_value=3.75, quantize_mode="dynamic")

    # TODO: uncomment following after ConvInteger s8 supported
    # def test_dynamic_mode_edge_s8s8(self):
    #     self.verify_quantize_with_pad_mode('edge', constant_value=None, quantize_mode='dynamic', activation_type=QuantType.QInt8,
    #                                        weight_type=QuantType.QInt8, extra_options={'ActivationSymmetric': True})

    # def test_dynamic_mode_reflect_s8s8(self):
    #     self.verify_quantize_with_pad_mode('reflect', constant_value=None, quantize_mode='dynamic', activation_type=QuantType.QInt8,
    #                                        weight_type=QuantType.QInt8, extra_options={'ActivationSymmetric': True})

    # def test_dynamic_mode_constant_default_s8s8(self):
    #     self.verify_quantize_with_pad_mode('constant', constant_value=None, quantize_mode='dynamic', activation_type=QuantType.QInt8,
    #                                        weight_type=QuantType.QInt8, extra_options={'ActivationSymmetric': True})

    # def test_dynamic_mode_constant_value_s8s8(self):
    #     self.verify_quantize_with_pad_mode('constant', constant_value=3.75, quantize_mode='dynamic', activation_type=QuantType.QInt8,
    #                                        weight_type=QuantType.QInt8, extra_options={'ActivationSymmetric': True})
    def construct_edge_case_model(
        self,
        output_model_path,
        conv_input_shape,
        conv_weight_shape,
        pad_input_shape,
        pad_mode,
        pad_dims,
        constant_value=None,
    ):
        #      (input)
        #          \
        #         Conv   (padding_value)
        #        /    \   /
        #   Identity   Pad
        #    /            \
        # (identity_out)  (output)
        rank = len(pad_input_shape)
        self.assertEqual(rank * 2, len(pad_dims))

        input_tensor = helper.make_tensor_value_info("input", TensorProto.FLOAT, conv_input_shape)
        conv_weight_arr = np.random.randint(-1, 2, conv_weight_shape).astype(np.float32)
        conv_weight_initializer = onnx.numpy_helper.from_array(conv_weight_arr, name="conv1_weight")
        conv_node = onnx.helper.make_node("Conv", ["input", "conv1_weight"], ["conv_output"], name="conv_node")

        identity_out = helper.make_tensor_value_info("identity_out", TensorProto.FLOAT, pad_input_shape)
        identity_node = helper.make_node("Identity", ["conv_output"], ["identity_out"], name="IdentityNode")

        pad_dims_initializer = helper.make_tensor("pad_dims", TensorProto.INT64, [2 * rank], pad_dims)
        output_shape = [sum(e) for e in list(zip(pad_input_shape, pad_dims[:rank], pad_dims[rank:], strict=False))]
        output_tensor = helper.make_tensor_value_info("output", TensorProto.FLOAT, output_shape)
        pad_inputs = ["conv_output", "pad_dims"]
        initializers = [conv_weight_initializer, pad_dims_initializer]
        constant_value_tensor = helper.make_tensor_value_info("padding_value", TensorProto.FLOAT, [1])
        pad_inputs.extend(["padding_value"])
        kwargs = {"mode": pad_mode} if pad_mode is not None else {}
        pad_node = helper.make_node("Pad", pad_inputs, ["output"], name="pad_node", **kwargs)

        graph = helper.make_graph(
            [conv_node, identity_node, pad_node],
            "TestOpQuantizerPad_test_model",
            [input_tensor, constant_value_tensor],
            [identity_out, output_tensor],
            initializer=initializers,
        )
        model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", 13)])
        model.ir_version = 7  # use stable onnx ir version
        onnx.save(model, output_model_path)

    def test_static_mode_constant_value_edge_case(self):
        self.verify_quantize_with_pad_mode(
            "constant", constant_value=0.1, quantize_mode="static", extra_options={"dual_feed": True}
        )

    @classmethod
    def construct_model_add_pad_add(
        cls,
        # Name of model input, i.e., "input" in the illustration graph below.
        name,
        # Name of model output.
        final_name,
        # model input shape.
        shape,
    ):
        # Graph implemented below is
        #  `name`, `name` -> Add -> "first_add_output"
        #  "first_add_output", "pads" -> Pad -> "pad_output"
        #  "pad_output", "pad_output" -> Add -> `final_name`
        # where `name` is the 2nd argument of this function,
        # `final_name` is the 3rd argument of this function,
        # and the rest lowercase strings are tensor names in the graph.

        input_name = name
        first_add_output_name = "first_add_output"
        pads_name = "pads"
        pad_output_name = "pad_output"
        second_add_output_name = final_name

        input_shape = shape
        input_rank = len(input_shape)

        input_tensor = helper.make_tensor_value_info(input_name, TensorProto.FLOAT, input_shape)

        first_add_node = helper.make_node(
            "Add",
            [input_name, input_name],
            [first_add_output_name],
            name="FirstAdd",
        )

        pads = [1, 2] * input_rank
        pads_initializer = helper.make_tensor(
            pads_name,
            TensorProto.INT64,
            # 1-D tensor of shape [2 * input_rank].
            [len(pads)],
            pads,
        )
        pad_node = helper.make_node(
            "Pad",
            [first_add_output_name, pads_name, ""],
            [pad_output_name],
            name="PadNode",
            mode="constant",
        )
        pad_output_shape = tuple(input_shape[i] + pads[i] + pads[i + input_rank] for i in range(input_rank))

        second_add_node = helper.make_node(
            "Add",
            [pad_output_name, pad_output_name],
            [second_add_output_name],
            name="SecondAdd",
        )

        output_tensor = helper.make_tensor_value_info(second_add_output_name, TensorProto.FLOAT, pad_output_shape)

        graph = helper.make_graph(
            [first_add_node, pad_node, second_add_node],
            "TestPadWithEmptyStringInput",
            [input_tensor],
            [output_tensor],
            initializer=[pads_initializer],
        )
        model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", 13)])
        model.ir_version = 7  # use stable onnx ir version

        return model

    def test_pad_with_empty_string_input_name(self):
        np.random.seed(108)
        model_fp32_path = "pad_with_empty_string_input_name_fp32.onnx"
        model_i8_path = "pad_with_empty_string_input_name_i8.onnx"

        shape = [
            3,
        ]
        name = "input"
        data_reader = self.input_feeds(
            1,
            {
                name: shape,
            },
        )

        model_fp32 = TestOpQuatizerPad.construct_model_add_pad_add(name=name, shape=shape, final_name="output")
        op_types = [n.op_type for n in model_fp32.graph.node]
        self.assertEqual(["Add", "Pad", "Add"], op_types)

        onnx.save(model_fp32, model_fp32_path)

        self.quantize_model(
            model_fp32_path,
            model_i8_path,
            data_reader=data_reader,
        )

        model_i8 = onnx.load(model_i8_path)
        print(model_i8)

        # Assert quantization really happens.
        op_types = [n.op_type for n in model_i8.graph.node]
        self.assertEqual(["QuantizeLinear", "QLinearAdd", "Pad", "QLinearAdd", "DequantizeLinear"], op_types)

        for node in model_i8.graph.node:
            # Examine no empty string flows to quantization process.
            # Previously, optional input specified by `""` in NodeProto.input
            # may cause phantom node to generate `"_quantized"` in quantization process.
            for name in itertools.chain(node.input, node.output):
                self.assertNotEqual(name, "")
                self.assertNotEqual(name, "_quantized")


class TestQDQPad(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls._tmp_model_dir = tempfile.TemporaryDirectory(prefix="ort.qdq.pad_")

        # Note: swap with the commented line if you want to see the models in local test dir.
        cls._tmp_dir_path = cls._tmp_model_dir.name
        # cls._tmp_dir_path = "."

    @classmethod
    def tearDownClass(cls):
        cls._tmp_model_dir.cleanup()

    def build_pad_model(
        self,
        mode: str,
        constant_value: float | None = None,
        opset: int = 21,
        float_type: onnx.TensorProto.DataType = onnx.TensorProto.FLOAT,
    ) -> onnx.ModelProto:
        num_pads_start = 1
        input_0 = onnx.helper.make_tensor_value_info("input_0", float_type, (3, 2))
        output_0 = onnx.helper.make_tensor_value_info("output_0", float_type, (3, 2 + num_pads_start))

        initializers = []
        pad_input_names = ["input_0"]
        attrs = {"mode": mode}

        pads_data = np.array([0, num_pads_start, 0, 0], dtype=np.int64)  # Pad one val at beginning of axis 1.
        if opset >= 11:
            initializers.append(onnx.numpy_helper.from_array(pads_data, "pads"))
            pad_input_names.append("pads")
        else:
            attrs["pads"] = pads_data.tolist()

        if mode == "constant" and constant_value is not None:
            if opset >= 11:
                initializers.append(onnx.helper.make_tensor("constant_value", float_type, [], [constant_value]))
                pad_input_names.append("constant_value")
            else:
                attrs["value"] = float(constant_value)

        pad_node = onnx.helper.make_node("Pad", pad_input_names, ["output_0"], name="Pad0", **attrs)

        graph = onnx.helper.make_graph(
            [pad_node],
            "PadFloat",
            [input_0],
            [output_0],
            initializer=initializers,
        )
        opset_imports = [onnx.helper.make_opsetid("", opset)]
        model = onnx.helper.make_model(graph, opset_imports=opset_imports)
        model = onnx.shape_inference.infer_shapes(model)
        onnx.checker.check_model(model, True)
        return model

    def test_qdq_pad_qparams(self):
        """
        Test that QDQ Pad has equal scale/zero-point for its input and output for certain configurations.
        """
        test_configs = [
            # Opset 21
            ("constant", None, 21, onnx.TensorProto.FLOAT),
            ("constant", None, 21, onnx.TensorProto.FLOAT16),
            ("constant", 0, 21, onnx.TensorProto.FLOAT),
            ("constant", 0, 21, onnx.TensorProto.FLOAT16),
            ("constant", 10.0, 21, onnx.TensorProto.FLOAT),
            ("constant", 10.0, 21, onnx.TensorProto.FLOAT16),
            ("reflect", None, 21, onnx.TensorProto.FLOAT),
            ("reflect", None, 21, onnx.TensorProto.FLOAT16),
            ("edge", None, 21, onnx.TensorProto.FLOAT),
            ("edge", None, 21, onnx.TensorProto.FLOAT16),
            ("wrap", None, 21, onnx.TensorProto.FLOAT),
            ("wrap", None, 21, onnx.TensorProto.FLOAT16),
            # Model with opset 10 will use pad of opset 2, which uses attributes instead of inputs.
            # Opset 10 Q/DQ ops don't support float16.
            ("constant", None, 10, onnx.TensorProto.FLOAT),
            ("constant", 0, 10, onnx.TensorProto.FLOAT),
            ("constant", 10.0, 10, onnx.TensorProto.FLOAT),
            ("reflect", None, 10, onnx.TensorProto.FLOAT),
            ("edge", None, 10, onnx.TensorProto.FLOAT),
        ]

        for pad_mode, constant_value, opset, float_type in test_configs:
            with self.subTest(pad_mode=pad_mode, constant_value=constant_value, opset=opset, float_type=float_type):
                label = f"_{pad_mode}_{constant_value}_opset{opset}_{onnx.TensorProto.DataType.Name(float_type)}"
                float_model_path = os.path.join(self._tmp_dir_path, f"pad{label}.float.onnx")
                qdq_model_path = os.path.join(self._tmp_dir_path, f"pad{label}.qdq.onnx")

                float_model = self.build_pad_model(pad_mode, constant_value, opset=opset, float_type=float_type)
                onnx.save_model(float_model, float_model_path)

                # Create a data reader
                np_dtype = onnx.helper.tensor_dtype_to_np_dtype(float_type)
                input_data_list = [
                    {"input_0": np.array([[1.0, 1.2], [2.3, 3.4], [4.5, 5.7]], dtype=np_dtype)},
                    {"input_0": np.array([[2.3, 3.4], [4.5, 5.7], [1.0, 1.2]], dtype=np_dtype)},
                ]
                data_reader = TestDataFeeds(input_data_list)

                # quantize model to QDQ
                quantize_static(
                    float_model_path,
                    qdq_model_path,
                    data_reader,
                    quant_format=QuantFormat.QDQ,
                    activation_type=QuantType.QUInt8,
                    weight_type=QuantType.QInt8,
                )

                expected_op_counts = {"DequantizeLinear": 2, "QuantizeLinear": 2, "Pad": 1}
                if constant_value is not None and opset >= 11:
                    expected_op_counts["DequantizeLinear"] += 1  # The constant padding value is quantized.
                check_op_type_count(self, qdq_model_path, **expected_op_counts)

                if pad_mode != "reflect":
                    # Do not check model correctness for 'reflect' mode because ONNX Runtime implementation does
                    # not match the ONNX reference implementation. See the following issue:
                    # https://github.com/microsoft/onnxruntime/issues/20801
                    data_reader.rewind()
                    check_model_correctness(self, float_model_path, qdq_model_path, data_reader.get_next())

                qdq_model = onnx.load_model(qdq_model_path)
                quant_output_same_as_input = False

                if pad_mode in ("reflect", "edge", "wrap"):
                    quant_output_same_as_input = True

                if pad_mode == "constant" and constant_value in (None, 0):
                    quant_output_same_as_input = True

                pad_node = next((node for node in qdq_model.graph.node if node.op_type == "Pad"), None)
                self.assertNotEqual(pad_node, None)
                self.assertEqual(pad_node.op_type, "Pad")

                # Get the parent and child nodes of the Pad and check that they are DQ/Q.
                consumers, producers = get_tensor_consumers_and_producers(qdq_model)
                input_dq_node = producers.get(pad_node.input[0], None)
                self.assertNotEqual(input_dq_node, None)
                self.assertEqual(input_dq_node.op_type, "DequantizeLinear")

                output_q_node = consumers.get(pad_node.output[0], [None])[0]
                self.assertNotEqual(output_q_node, None)
                self.assertEqual(output_q_node.op_type, "QuantizeLinear")

                # Check that the Pad's input DQ uses the same scale/zp as the Pad's output Q.
                if quant_output_same_as_input:
                    self.assertEqual(input_dq_node.input[1], output_q_node.input[1])  # Same scale
                    self.assertEqual(input_dq_node.input[2], output_q_node.input[2])  # Same zero-point
                else:
                    self.assertNotEqual(input_dq_node.input[1], output_q_node.input[1])
                    self.assertNotEqual(input_dq_node.input[2], output_q_node.input[2])


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
from __future__ import annotations

import logging
import pathlib
import unittest

from testfixtures import LogCapture

from ..usability_checker import analyze_model

# example usage from <ort root>/tools/python
# python -m unittest util/mobile_helpers/test/test_usability_checker.py
# NOTE: at least on Windows you must use that as the working directory for all the imports to be happy

script_dir = pathlib.Path(__file__).parent
ort_root = script_dir.parents[4]

skip_optimize = False


def _create_logger():
    logger = logging.getLogger("default")
    logger.setLevel(logging.DEBUG)
    return logger


class TestAnalyzer(unittest.TestCase):
    def test_mnist(self):
        """
        Test MNIST which should be fully covered by both NNAPI and CoreML as is.
        :return:
        """
        with LogCapture() as log_capture:
            logger = _create_logger()
            model_path = ort_root / "onnxruntime" / "test" / "testdata" / "mnist.onnx"
            analyze_model(model_path, skip_optimize, logger)

            # print(log_capture)
            log_capture.check_present(
                ("default", "INFO", "1 partitions with a total of 8/8 nodes can be handled by the NNAPI EP."),
                ("default", "INFO", "Model should perform well with NNAPI as is: YES"),
                (
                    "default",
                    "INFO",
                    "1 partitions with a total of 8/8 nodes can be handled by the CoreML NeuralNetwork EP.",
                ),
                ("default", "INFO", "Model should perform well with CoreML NeuralNetwork as is: YES"),
                (
                    "default",
                    "INFO",
                    "1 partitions with a total of 8/8 nodes can be handled by the CoreML MLProgram EP.",
                ),
                ("default", "INFO", "Model should perform well with CoreML MLProgram as is: YES"),
            )

    def test_scan_model(self):
        """
        Test a Speech model where all the top level nodes are Scan. We want to make sure nodes in subgraphs are counted.
        """
        with LogCapture() as log_capture:
            logger = _create_logger()
            # mnist - should have perfect coverage
            model_path = ort_root / "onnxruntime" / "test" / "testdata" / "scan_1.onnx"
            analyze_model(model_path, skip_optimize, logger)

            # print(log_capture)
            log_capture.check_present(
                ("default", "INFO", "4 partitions with a total of 72/76 nodes can be handled by the NNAPI EP."),
                ("default", "INFO", "72 nodes are in 4 subgraphs. Check EP as to whether subgraphs are supported."),
                ("default", "INFO", "Model should perform well with NNAPI as is: NO"),
                (
                    "default",
                    "INFO",
                    "4 partitions with a total of 60/76 nodes can be handled by the CoreML NeuralNetwork EP.",
                ),
                ("default", "INFO", "Model should perform well with CoreML NeuralNetwork as is: NO"),
                (
                    "default",
                    "INFO",
                    "12 partitions with a total of 24/76 nodes can be handled by the CoreML MLProgram EP.",
                ),
                ("default", "INFO", "Model should perform well with CoreML MLProgram as is: NO"),
            )

    def test_dynamic_shape(self):
        """
        Test a model with dynamic input shape and supported op.
        If we make the shape fixed it should report it will run well with NNAPI/CoreML.
        """
        with LogCapture() as log_capture:
            logger = _create_logger()
            model_path = ort_root / "onnxruntime" / "test" / "testdata" / "abs_free_dimensions.onnx"
            analyze_model(model_path, skip_optimize, logger)

            # print(log_capture)
            log_capture.check_present(
                ("default", "INFO", "0 partitions with a total of 0/1 nodes can be handled by the NNAPI EP."),
                ("default", "INFO", "Model should perform well with NNAPI as is: NO"),
                ("default", "INFO", "Model should perform well with NNAPI if modified to have fixed input shapes: YES"),
                (
                    "default",
                    "INFO",
                    "0 partitions with a total of 0/1 nodes can be handled by the CoreML MLProgram EP.",
                ),
                ("default", "INFO", "CoreML MLProgram cannot run any nodes in this model."),
                ("default", "INFO", "Model should perform well with CoreML MLProgram as is: NO"),
                (
                    "default",
                    "INFO",
                    "Model should perform well with CoreML MLProgram if modified to have fixed input shapes: NO",
                ),
            )

    def test_multi_partitions(self):
        """
        Test a model that breaks into too many partitions to be recommended for use with NNAPI/CoreML
        """
        with LogCapture() as log_capture:
            logger = _create_logger()
            model_path = ort_root / "onnxruntime" / "test" / "testdata" / "gh_issue_9671.onnx"
            analyze_model(model_path, skip_optimize, logger)

            # print(log_capture)
            log_capture.check_present(
                ("default", "INFO", "3 partitions with a total of 22/50 nodes can be handled by the NNAPI EP."),
                ("default", "INFO", "\tPartition sizes: [13, 2, 7]"),
                (
                    "default",
                    "INFO",
                    "\tUnsupported ops: ai.onnx:ReduceProd,ai.onnx:ReduceSum,ai.onnx:Shape",
                ),
                (
                    "default",
                    "INFO",
                    "NNAPI is not recommended with this model as there are 3 partitions "
                    "covering 44.0% of the nodes in the model. "
                    "This will most likely result in worse performance than just using the CPU EP.",
                ),
                (
                    "default",
                    "INFO",
                    "4 partitions with a total of 20/50 nodes can be handled by the CoreML NeuralNetwork EP.",
                ),
                ("default", "INFO", "\tPartition sizes: [11, 3, 5, 1]"),
            )



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import numpy as np
import onnx


def subgraph_1d_const_input_dq(inputs, initializers, nodes) -> str:
    """
    Creates mul_weight -> DQ. mul_weight is a constant of rank 1.
    """
    mul_weight_i8_data = np.array([1, 2, 3], dtype=np.int8)
    mul_weight = onnx.numpy_helper.from_array(mul_weight_i8_data, "mul_weight")
    initializers.append(mul_weight)

    dq_output_name = "mul_input_1"
    nodes.append(
        onnx.helper.make_node(
            "DequantizeLinear",
            ["mul_weight", "mul_weight_scales", "mul_weight_zps"],
            [dq_output_name],
            name="dq_mul_input_1",
            axis=0,
        )
    )

    return dq_output_name


def subgraph_1d_input_dq(inputs, initializers, nodes) -> str:
    """
    Creates input1 -> DQ. input1 is a graph input of rank 1.
    """
    input1_shape = (3,)
    inputs.append(onnx.helper.make_tensor_value_info("input1", onnx.TensorProto.INT8, input1_shape))

    dq_output_name = "mul_input_1"
    nodes.append(
        onnx.helper.make_node(
            "DequantizeLinear",
            ["input1", "mul_weight_scales", "mul_weight_zps"],
            [dq_output_name],
            name="dq_mul_input_1",
            axis=0,
        )
    )

    return dq_output_name


def subgraph_4d_input_squeeze_dq(inputs, initializers, nodes) -> str:
    """
    Creates input1 -> Squeeze -> DQ. input1 is a graph input of rank 4.
    """
    input1_shape = (1, 1, 1, 3)
    inputs.append(onnx.helper.make_tensor_value_info("input1", onnx.TensorProto.INT8, input1_shape))

    axes_data = np.array([0, 1, 2], dtype=np.int64)
    initializers.append(onnx.numpy_helper.from_array(axes_data, "axes_const"))
    nodes.append(onnx.helper.make_node("Squeeze", ["input1", "axes_const"], ["squeeze_out"], name="squeeze_node"))

    dq_output_name = "mul_input_1"
    nodes.append(
        onnx.helper.make_node(
            "DequantizeLinear",
            ["squeeze_out", "mul_weight_scales", "mul_weight_zps"],
            [dq_output_name],
            name="dq_mul_input_1",
            axis=0,
        )
    )

    return dq_output_name


def subgraph_4d_input_transpose_dq(inputs, initializers, nodes) -> str:
    """
    Creates input1 -> Transpose -> DQ. input1 is a graph input of rank 4.
    """
    input1_shape = (1, 3, 1, 1)
    inputs.append(onnx.helper.make_tensor_value_info("input1", onnx.TensorProto.INT8, input1_shape))

    perm = [0, 2, 3, 1]  # To channel-last
    nodes.append(onnx.helper.make_node("Transpose", ["input1"], ["tp_out_"], perm=perm, name="transpose_"))

    dq_output_name = "mul_input_1"
    nodes.append(
        onnx.helper.make_node(
            "DequantizeLinear",
            ["tp_out_", "mul_weight_scales", "mul_weight_zps"],
            [dq_output_name],
            name="dq_mul_input_1",
            axis=-1,
        )
    )

    return dq_output_name


def make_model(model_path: str, build_mul_input_1_subgraph):
    """
    Creates a QDQ model with a per-axis DQ input that is Unsqueezed and Transposed by the Transpose optimizer.
    """
    input0_shape = (1, 3, 4, 4)

    inputs = [onnx.helper.make_tensor_value_info("input0", onnx.TensorProto.FLOAT, input0_shape)]
    outputs = [onnx.helper.make_tensor_value_info("output0", onnx.TensorProto.FLOAT, None)]

    mul_weight_scales_data = np.array([1.0, 1.0, 1.0], dtype=np.float32)
    mul_weight_zps_data = np.array([0, 0, 0], dtype=np.int8)

    initializers = [
        onnx.numpy_helper.from_array(np.array(1.0, dtype=np.float32), "scale_1"),
        onnx.numpy_helper.from_array(np.array(128, dtype=np.uint8), "zp_128"),
        onnx.numpy_helper.from_array(np.array(1.0 / 255.0, dtype=np.float32), "scale_inv_255"),
        onnx.numpy_helper.from_array(np.array(0, dtype=np.uint8), "zp_0"),
        onnx.numpy_helper.from_array(mul_weight_scales_data, "mul_weight_scales"),
        onnx.numpy_helper.from_array(mul_weight_zps_data, "mul_weight_zps"),
    ]
    nodes = []

    # Transpose to channel-last
    tp0_node = onnx.helper.make_node("Transpose", ["input0"], ["tp0_out"], name="tp0_node", perm=(0, 2, 3, 1))
    nodes.append(tp0_node)

    # Q_0
    q0_node = onnx.helper.make_node("QuantizeLinear", ["tp0_out", "scale_1", "zp_128"], ["q0_out"], name="q0_node")
    nodes.append(q0_node)

    # DQ_0
    dq0_node = onnx.helper.make_node("DequantizeLinear", ["q0_out", "scale_1", "zp_128"], ["dq0_out"], name="dq0_node")
    nodes.append(dq0_node)

    # Sigmoid
    sigmoid_node = onnx.helper.make_node("Sigmoid", ["dq0_out"], ["sigmoid_out"], name="sigmoid_node")
    nodes.append(sigmoid_node)

    # Q_1
    q1_node = onnx.helper.make_node(
        "QuantizeLinear", ["sigmoid_out", "scale_inv_255", "zp_0"], ["q1_out"], name="q1_node"
    )
    nodes.append(q1_node)

    # DQ_1
    dq1_node = onnx.helper.make_node(
        "DequantizeLinear", ["q1_out", "scale_inv_255", "zp_0"], ["dq1_out"], name="dq1_node"
    )
    nodes.append(dq1_node)

    # DQ for mul input[1]
    mul_input_1_name = build_mul_input_1_subgraph(inputs, initializers, nodes)

    # Mul
    mul_node = onnx.helper.make_node("Mul", ["dq1_out", mul_input_1_name], ["mul_out"], name="mul_node")
    nodes.append(mul_node)

    # Q_2
    q2_node = onnx.helper.make_node("QuantizeLinear", ["mul_out", "scale_inv_255", "zp_0"], ["q2_out"], name="q2_node")
    nodes.append(q2_node)

    # DQ_2
    dq2_node = onnx.helper.make_node(
        "DequantizeLinear", ["q2_out", "scale_inv_255", "zp_0"], ["dq2_out"], name="dq2_node"
    )
    nodes.append(dq2_node)

    # Transpose to channel-first
    tp1_node = onnx.helper.make_node("Transpose", ["dq2_out"], ["output0"], name="tp1_node", perm=(0, 3, 1, 2))
    nodes.append(tp1_node)

    graph = onnx.helper.make_graph(
        nodes,
        "transpose_opt_unsqueeze_dq_axis",
        inputs,
        outputs,
        initializer=initializers,
    )
    opset_imports = [
        onnx.helper.make_opsetid("", 19),
    ]
    qdq_model = onnx.helper.make_model(graph, opset_imports=opset_imports)

    print("[INFO]: Running onnx.checker on qdq model")
    qdq_model = onnx.shape_inference.infer_shapes(qdq_model)
    onnx.checker.check_model(qdq_model, True)

    print(f"[INFO]: Saving {model_path}")
    onnx.save_model(qdq_model, model_path)


if __name__ == "__main__":
    make_model(
        "transpose_optimizer_qdq_fixup_unsqueeze_per_axis_dq.onnx",
        subgraph_1d_input_dq,
    )
    make_model(
        "transpose_optimizer_in_place_transpose_unsqueeze_per_axis_dq.onnx",
        subgraph_1d_const_input_dq,
    )
    make_model(
        "transpose_optimizer_cancel_squeeze_per_axis_dq.onnx",
        subgraph_4d_input_squeeze_dq,
    )
    make_model(
        "transpose_optimizer_cancel_transpose_per_axis_dq.onnx",
        subgraph_4d_input_transpose_dq,
    )



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import contextlib
import copy
import logging
import os
from abc import ABC, abstractmethod
from typing import Any

import numpy as np
import onnx

import onnxruntime.training.onnxblock._graph_utils as _graph_utils
import onnxruntime.training.onnxblock.model_accessor as accessor


class Block(ABC):
    """Base class for all building blocks that can be stacked on top of each other.

    All blocks that want to manipulate the model must subclass this class. The subclass's
    implementation of the build method must return the names of the intermediate outputs from
    the block.

    The subclass's implementation of the build method must manipulate the base model as it deems fit,
    but the manipulated model must be valid (as deemed by the onnx checker).

    Attributes:
        base (onnx.ModelProto): The base model that the subclass can manipulate.
    """

    def __init__(self, temp_file_name="temp.onnx"):
        if os.path.isabs(temp_file_name):
            raise RuntimeError("Please pass in a relative path for the temp_file_name.")
        self.base = None
        self.temp_onnx_file_path = os.path.join(os.getcwd(), temp_file_name)
        # onnx.save location parameter requires a relative path to the model path
        self.temp_external_data_file_name = temp_file_name + ".data"

    @abstractmethod
    def build(self, *args, **kwargs):
        """Customize the model by stacking up blocks on top of the inputs to this function.

        This method must be overridden by the subclass.
        """

    def __call__(self, *args, **kwargs):
        """Calls the subclass's build method and runs validation on top."""

        self.base = accessor._GLOBAL_ACCESSOR.model

        logging.debug("Building block: %s", self.__class__.__name__)

        output = self.build(*args, **kwargs)

        if accessor._GLOBAL_ACCESSOR.has_path:
            # `save` will destructively access any external data
            copied_model = copy.deepcopy(accessor._GLOBAL_ACCESSOR.model)
            onnx.save(
                copied_model,
                self.temp_onnx_file_path,
                save_as_external_data=True,
                all_tensors_to_one_file=True,
                location=self.temp_external_data_file_name,
            )

            onnx.checker.check_model(self.temp_onnx_file_path, True)
        else:
            onnx.checker.check_model(self.base, True)

        return output

    def infer_shapes_on_base(self):
        """
        Performs shape inference on the global model. If a path was used, then uses the
        infer_shapes_path API to support models with external data.

        Returns the shape-inferenced ModelProto.
        """
        if accessor._GLOBAL_ACCESSOR.has_path:
            onnx.save(
                accessor._GLOBAL_ACCESSOR.model,
                self.temp_onnx_file_path,
                save_as_external_data=True,
                all_tensors_to_one_file=True,
                location=self.temp_external_data_file_name,
            )

            onnx.shape_inference.infer_shapes_path(self.temp_onnx_file_path)
            # shape inferenced model is saved to original path
            model = onnx.load(self.temp_onnx_file_path)

            return model
        else:
            return onnx.shape_inference.infer_shapes(accessor._GLOBAL_ACCESSOR.model)

    def __del__(self):
        # since the ModelProto does not store the external data parameters themselves, just the metadata
        # for where the external data can be found, we retain the external data files for the intermediate
        # calls until the Block no longer needs to be used.
        if os.path.exists(self.temp_onnx_file_path):
            os.remove(self.temp_onnx_file_path)
            # get absolute path for the external data file
            external_data_file_path = os.path.join(
                os.path.dirname(self.temp_onnx_file_path), self.temp_external_data_file_name
            )
            if os.path.exists(external_data_file_path):
                os.remove(external_data_file_path)


class _BinaryOp(Block):
    def __init__(self, op_name):
        super().__init__()
        self._op_name = op_name

    def build(self, input_name1, input_name2):
        # get the model to manipulate
        onnx_model = self.base

        # Assert that the op name is not empty
        if not self._op_name:
            raise RuntimeError("Unknown op name. Please override _op_name")

        # create the graph node for sub
        node_input_names = [input_name1, input_name2]
        node_output_name = _graph_utils.generate_graph_name(f"{self._op_name.lower()}_output")
        node_output_names = [node_output_name]
        node = onnx.helper.make_node(
            self._op_name,
            node_input_names,
            node_output_names,
            name=_graph_utils.generate_graph_name(self._op_name),
        )
        onnx_model.graph.node.append(node)

        return node_output_name


class Add(_BinaryOp):
    """Adds Add node to an onnx model."""

    def __init__(self):
        super().__init__("Add")


class Sub(_BinaryOp):
    """Adds Sub node to an onnx model."""

    def __init__(self):
        super().__init__("Sub")


class Mul(_BinaryOp):
    """Adds Mul node to an onnx model."""

    def __init__(self):
        super().__init__("Mul")


class Div(_BinaryOp):
    """Adds Div node to an onnx model."""

    def __init__(self):
        super().__init__("Div")


class Pow(Block):
    """Adds Pow node to the onnx model."""

    def __init__(self, exponent):
        super().__init__()

        self._exponent = exponent

    def build(self, pow_input_name):
        # get the model to manipulate
        onnx_model = self.base

        # create the graph initializer for the exponent
        pow_node_exponent_name = _graph_utils.generate_graph_name("pow_exponent")
        onnx_model.graph.initializer.append(
            onnx.helper.make_tensor(pow_node_exponent_name, onnx.TensorProto.FLOAT, [1], [self._exponent])
        )

        # create the graph node for pow
        pow_node_input_names = [pow_input_name, pow_node_exponent_name]
        pow_node_output_name = _graph_utils.generate_graph_name("pow_output")
        pow_node_output_names = [pow_node_output_name]
        pow_node = onnx.helper.make_node(
            "Pow",
            pow_node_input_names,
            pow_node_output_names,
            name=_graph_utils.generate_graph_name("Pow"),
        )
        onnx_model.graph.node.append(pow_node)

        return pow_node_output_name


class _UnaryOp(Block):
    """Base class for all nodes that take in a single argument."""

    def __init__(self, op_name, **attributes):
        super().__init__()
        self._op_name = op_name
        self._attributes = attributes

    def build(self, input_name):
        # get the model to manipulate
        onnx_model = self.base

        # Assert that the op name is not empty
        if not self._op_name:
            raise RuntimeError("Unknown op name. Please override _op_name")

        # create the graph node for this unary op
        node_input_names = [input_name]
        node_output_name = _graph_utils.generate_graph_name(f"{self._op_name.lower()}_output")
        node_output_names = [node_output_name]
        node = onnx.helper.make_node(
            self._op_name,
            node_input_names,
            node_output_names,
            _graph_utils.generate_graph_name(self._op_name),
            **self._attributes,
        )
        onnx_model.graph.node.append(node)

        return node_output_name


class ReduceMean(_UnaryOp):
    """Adds ReduceMean node to the onnx model."""

    def __init__(self, keepdims=True):
        super().__init__("ReduceMean", keepdims=keepdims)


class ReduceSum(_UnaryOp):
    """Adds ReduceSum node to the onnx model."""

    def __init__(self, keepdims=True):
        super().__init__("ReduceSum", keepdims=keepdims)


class Sigmoid(_UnaryOp):
    """Adds Sigmoid node to the onnx model."""

    def __init__(self):
        super().__init__("Sigmoid")


class Log(_UnaryOp):
    """Adds Log node to the onnx model."""

    def __init__(self):
        super().__init__("Log")


class Neg(_UnaryOp):
    """Adds Neg node to the onnx model."""

    def __init__(self):
        super().__init__("Neg")


class Abs(_UnaryOp):
    """Adds Abs node to the onnx model."""

    def __init__(self):
        super().__init__("Abs")


class Constant(Block):
    """Creates a float initializer and adds it to the onnx model."""

    # TODO: Add ability to add all sorts of initializers (not just floats).

    def __init__(self, value_float):
        super().__init__()

        self._value = value_float

    def build(self):
        # create the graph initializer for the exponent
        initializer_name = _graph_utils.generate_graph_name("initializer")
        self.base.graph.initializer.append(
            onnx.helper.make_tensor(initializer_name, onnx.TensorProto.FLOAT, [1], [self._value])
        )
        return initializer_name


class SequenceConstruct(Block):
    """Adds SequenceConstruct node to the onnx model."""

    def build(self, *sequence_input_names):
        sc_node_input_names = list(sequence_input_names)
        sc_node_output_name = _graph_utils.generate_graph_name("sequenceconstruct_output")
        sc_node_output_names = [sc_node_output_name]
        sc_node = onnx.helper.make_node(
            "SequenceConstruct",
            sc_node_input_names,
            sc_node_output_names,
            _graph_utils.generate_graph_name("SequenceConstruct"),
        )
        self.base.graph.node.append(sc_node)

        return sc_node_output_name


class ReduceAllL2(Block):
    """Adds ReduceAllL2 node to the onnx model.

    ReduceAllL2 is a part of the com.microsoft domain and might not be accessible outside this domain.
    """

    def build(self, *reduce_node_input_names):
        reduce_node_input_names = list(reduce_node_input_names)
        reduce_node_output_name = _graph_utils.generate_graph_name("reducealll2_output")
        reduce_node_output_names = [reduce_node_output_name]
        reduce_node = onnx.helper.make_node(
            "ReduceAllL2",
            reduce_node_input_names,
            reduce_node_output_names,
            _graph_utils.generate_graph_name("ReduceAllL2"),
            domain="com.microsoft",
        )
        self.base.graph.node.append(reduce_node)
        # TODO: register shape inference with onnx
        self.base.graph.value_info.append(
            onnx.helper.make_tensor_value_info(reduce_node_output_name, onnx.TensorProto.FLOAT, [1])
        )

        return reduce_node_output_name


class Clip(Block):
    """Adds Clip node to the onnx model."""

    def __init__(self, clip_min=None, clip_max=None):
        super().__init__()

        self._min = clip_min
        self._max = clip_max

    def build(self, clip_input_name):
        # get the model to manipulate
        onnx_model = self.base

        # create the graph initializer for the clip min
        clip_node_min_name = ""
        if self._min is not None:
            clip_node_min_name = _graph_utils.generate_graph_name("clip_min")
            onnx_model.graph.initializer.append(
                onnx.helper.make_tensor(clip_node_min_name, onnx.TensorProto.FLOAT, [1], [self._min])
            )

        # create the graph initializer for the clip max
        clip_node_max_name = ""
        if self._max is not None:
            clip_node_max_name = _graph_utils.generate_graph_name("clip_max")
            onnx_model.graph.initializer.append(
                onnx.helper.make_tensor(clip_node_max_name, onnx.TensorProto.FLOAT, [1], [self._max])
            )

        # create the graph node for this clip node
        clip_node_input_names = [
            clip_input_name,
            clip_node_min_name,
            clip_node_max_name,
        ]
        clip_node_output_name = _graph_utils.generate_graph_name("clip_output")
        clip_node_output_names = [clip_node_output_name]
        clip_node = onnx.helper.make_node(
            "Clip",
            clip_node_input_names,
            clip_node_output_names,
            _graph_utils.generate_graph_name("Clip"),
        )
        onnx_model.graph.node.append(clip_node)

        return clip_node_output_name


class PassThrough(Block):
    """A pass through block that returns the inputs without making any changes to onnx."""

    def build(self, *input_names):
        return input_names


class InputLike(Block):
    """Add an input to the onnx model like the graph input/output associated with the given name"""

    def __init__(self, like: str):
        """Create an input like the graph input/output associated with the given name

        Args:
            like (str): The name of the graph input/output to clone the type and shape from
        """
        super().__init__()

        self._like = like

    def build(self, input_name: str | None = None):
        cloned_input = None
        with contextlib.suppress(LookupError):
            # Suppress LookupError because we want to try to get the input from the output if it's not found in the inputs
            cloned_input = copy.deepcopy(_graph_utils.get_input_from_input_name(self.base, self._like))

        if cloned_input is None:
            with contextlib.suppress(LookupError):
                # Suppress LookupError because we deal with the case where no input or output was found later.
                cloned_input = copy.deepcopy(_graph_utils.get_output_from_output_name(self.base, self._like))

        if cloned_input is None:
            raise LookupError(f"Could not find input or output with name {self._like}")

        cloned_input.name = input_name or _graph_utils.generate_graph_name("input")
        self.base.graph.input.append(cloned_input)

        return cloned_input.name


class LabelEncoder(Block):
    def __init__(
        self,
        default_float: float = 0.0,
        default_int64: int = -1,
        default_string: str = "_Unused",
        keys_floats: list[float] | None = None,
        keys_int64s: list[int] | None = None,
        keys_strings: list[str] | None = None,
        values_floats: list[float] | None = None,
        values_int64s: list[int] | None = None,
        values_strings: list[str] | None = None,
    ):
        super().__init__()

        self._attributes = {
            "default_float": default_float,
            "default_int64": default_int64,
            "default_string": default_string,
        }

        def _add_attributes(names: list[str], values: list[Any]):
            for name, value in zip(names, values, strict=False):
                if value is not None:
                    self._attributes[name] = value

        _add_attributes(
            ["keys_floats", "keys_int64s", "keys_strings", "values_floats", "values_int64s", "values_strings"],
            [keys_floats, keys_int64s, keys_strings, values_floats, values_int64s, values_strings],
        )

    def build(self, label_encoder_input_name: str):
        label_encoder_output_name = _graph_utils.generate_graph_name("label_encoder.output")
        label_encoder_node = onnx.helper.make_node(
            "LabelEncoder",
            [label_encoder_input_name],
            [label_encoder_output_name],
            _graph_utils.generate_graph_name("LabelEncoder"),
            domain="ai.onnx.ml",
            **self._attributes,
        )
        self.base.graph.node.append(label_encoder_node)

        return label_encoder_output_name


class Cast(Block):
    def __init__(self, to: onnx.TensorProto.DataType):
        super().__init__()

        self._to = to

    def build(self, cast_input_name: str):
        cast_output_name = _graph_utils.generate_graph_name("cast.output")
        cast_node = onnx.helper.make_node(
            "Cast",
            [cast_input_name],
            [cast_output_name],
            _graph_utils.generate_graph_name("Cast"),
            to=self._to,
        )
        self.base.graph.node.append(cast_node)

        return cast_output_name


class Linear(Block):
    def __init__(self, in_features, out_features, bias=True, alpha=1.0, beta=1.0):
        super().__init__()

        self._in_features = in_features
        self._bias = bias
        self._out_features = out_features
        self._alpha = alpha
        self._beta = beta

    def build(self, linear_input_name: str):
        # Weight initializer
        linear_node_weight_name = _graph_utils.generate_graph_name("linear.weight")

        self.base.graph.initializer.append(
            onnx.numpy_helper.from_array(
                np.random.randn(self._in_features, self._out_features).astype(np.float32), linear_node_weight_name
            )
        )

        linear_node_input_names = [linear_input_name, linear_node_weight_name]

        # Bias initializer
        if self._bias:
            linear_node_bias_name = _graph_utils.generate_graph_name("linear.bias")
            self.base.graph.initializer.append(
                onnx.numpy_helper.from_array(
                    np.random.randn(self._out_features).astype(np.float32), linear_node_bias_name
                )
            )
            linear_node_input_names.append(linear_node_bias_name)

        linear_node_output_name = _graph_utils.generate_graph_name("linear.output")
        linear_node_output_names = [linear_node_output_name]
        linear_node = onnx.helper.make_node(
            "Gemm",
            linear_node_input_names,
            linear_node_output_names,
            _graph_utils.generate_graph_name("linear"),
            alpha=self._alpha,
            beta=self._beta,
        )

        self.base.graph.node.append(linear_node)

        return linear_node_output_name



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import os

import cntk as C
import numpy as np
import onnx
from onnx import numpy_helper

model_file = "model.onnx"
data_dir = "test_data_set_0"


def SaveTensorProto(file_path, variable, data, name):  # noqa: N802
    # ONNX input shape always has sequence axis as the first dimension, if sequence axis exists
    if len(variable.dynamic_axes) == 2:
        data = data.transpose((1, 0, *tuple(range(2, len(data.shape)))))
    tp = numpy_helper.from_array(data, name if name else variable.uid)
    onnx.save_tensor(tp, file_path)


def SaveData(test_data_dir, prefix, variables, data_list, name_replacements=None):  # noqa: N802
    if isinstance(data_list, np.ndarray):
        data_list = [data_list]
    for (i, d), v in zip(enumerate(data_list), variables, strict=False):
        SaveTensorProto(
            os.path.join(test_data_dir, f"{prefix}_{i}.pb"),
            v,
            d,
            name_replacements[v.uid] if name_replacements else None,
        )


def Save(dir, func, feed, outputs):  # noqa: N802
    if not os.path.exists(dir):
        os.makedirs(dir)
    onnx_file = os.path.join(dir, model_file)
    func.save(onnx_file, C.ModelFormat.ONNX)

    # onnx model may have different name for RNN initial states as inputs
    cntk_to_actual_names = {}
    model = onnx.load(onnx_file)
    for actual_input in model.graph.input:
        actual_input_name = actual_input.name
        for cntk_input in func.arguments:
            cntk_name = cntk_input.uid
            if actual_input_name.startswith(cntk_name):
                cntk_to_actual_names[cntk_name] = actual_input_name

    if type(feed) is not dict:
        feed = {func.arguments[0]: feed}

    if type(outputs) is not dict:
        outputs = {func.outputs[0]: outputs}

    test_data_dir = os.path.join(dir, data_dir)
    if not os.path.exists(test_data_dir):
        os.makedirs(test_data_dir)

    SaveData(
        test_data_dir,
        "input",
        func.arguments,
        [feed[var] for var in func.arguments],
        cntk_to_actual_names,
    )
    SaveData(test_data_dir, "output", func.outputs, [outputs[var] for var in func.outputs])


def GenSimple():  # noqa: N802
    x = C.input_variable(
        (
            1,
            3,
        )
    )  # TODO: fix CNTK exporter bug with shape (3,)
    y = C.layers.Embedding(2)(x) + C.parameter((-1,))
    data_x = np.random.rand(1, *x.shape).astype(np.float32)
    data_y = y.eval(data_x)
    Save("test_simple", y, data_x, data_y)


def GenSharedWeights():  # noqa: N802
    x = C.input_variable(
        (
            1,
            3,
        )
    )
    y = C.layers.Embedding(2)(x)
    y = y + y.parameters[0]
    data_x = np.random.rand(1, *x.shape).astype(np.float32)
    data_y = y.eval(data_x)
    Save("test_shared_weights", y, data_x, data_y)


def GenSimpleMNIST():  # noqa: N802
    input_dim = 784
    num_output_classes = 10
    num_hidden_layers = 1
    hidden_layers_dim = 200

    feature = C.input_variable(input_dim, np.float32)

    scaled_input = C.element_times(C.constant(0.00390625, shape=(input_dim,)), feature)

    z = C.layers.Sequential(
        [
            C.layers.For(
                range(num_hidden_layers),
                lambda i: C.layers.Dense(hidden_layers_dim, activation=C.relu),
            ),
            C.layers.Dense(num_output_classes),
        ]
    )(scaled_input)

    model = C.softmax(z)

    data_feature = np.random.rand(1, *feature.shape).astype(np.float32)
    data_output = model.eval(data_feature)
    Save("test_simpleMNIST", model, data_feature, data_output)


def GenMatMul_1k():  # noqa: N802
    feature = C.input_variable(
        (
            1024,
            1024,
        ),
        np.float32,
    )
    model = C.times(feature, C.parameter((1024, 1024), init=C.glorot_uniform()))

    data_feature = np.random.rand(1, *feature.shape).astype(np.float32)
    data_output = model.eval(data_feature)
    Save("test_MatMul_1k", model, data_feature, data_output)


def LSTM(cell_dim, use_scan=True):  # noqa: N802
    # we now create an LSTM_cell function and call it with the input and placeholders
    LSTM_cell = C.layers.LSTM(cell_dim)  # noqa: N806

    @C.Function
    def func(dh, dc, input):
        LSTM_func = LSTM_cell(dh, dc, input)  # noqa: N806
        if use_scan:
            LSTM_func_root = C.as_composite(LSTM_func.outputs[0].owner.block_root)  # noqa: N806
            args = LSTM_func_root.arguments
            LSTM_func = LSTM_func_root.clone(  # noqa: N806
                C.CloneMethod.share, {args[0]: input, args[1]: dh, args[2]: dc}
            )
        return LSTM_func

    return func


def GenLSTMx4(use_scan):  # noqa: N802
    feature = C.sequence.input_variable((128,), np.float32)
    lstm1 = C.layers.Recurrence(LSTM(512, use_scan))(feature)
    lstm2_fw = C.layers.Recurrence(LSTM(512, use_scan))(lstm1)
    lstm2_bw = C.layers.Recurrence(LSTM(512, use_scan), go_backwards=True)(lstm1)
    lstm2 = C.splice(lstm2_fw, lstm2_bw, axis=0)
    lstm3_fw = C.layers.Recurrence(LSTM(512, use_scan))(lstm2)
    lstm3_bw = C.layers.Recurrence(LSTM(512, use_scan), go_backwards=True)(lstm2)
    lstm3 = C.splice(lstm3_fw, lstm3_bw, axis=0)
    lstm4 = C.layers.Recurrence(LSTM(512, use_scan))(lstm3)
    model = lstm4

    postfix = "Scan" if use_scan else "LSTM"

    data_feature = np.random.rand(1, 64, 128).astype(np.float32)
    data_output = np.asarray(model.eval(data_feature))
    Save("test_LSTMx4_" + postfix, model, data_feature, data_output)


def GenScan():  # noqa: N802
    np.random.seed(0)
    feature = C.sequence.input_variable((3,), np.float32)
    model = C.layers.For(range(4), lambda: C.layers.Recurrence(LSTM(2, use_scan=True)))(feature)

    data_feature = np.random.rand(2, 5, 3).astype(np.float32)
    data_output = np.asarray(model.eval(data_feature))

    Save("test_Scan", model, data_feature, data_output)

    # Currently CNTK only outputs batch == 1, do some editing
    in_mp = onnx.load("test_Scan/model.onnx")
    out_mp = onnx.ModelProto()
    out_mp.CopyFrom(in_mp)
    out_mp.graph.ClearField("initializer")

    # change LSTM init_c/h into inputs to support truncated sequence
    # as batch dimension is unknown on those data when building model
    # note here we assume init_c/h starts from 0
    # if not the case, user need to manually broadcast it for feed
    num_inputs = 1
    for i in in_mp.graph.initializer:
        if i.name.startswith("Constant"):
            shape = i.dims
            shape[0] = 2
            aa = np.zeros(shape, dtype=np.float32)
            tp = numpy_helper.from_array(aa, i.name)
            with open("test_Scan/test_data_set_0/input_" + str(num_inputs) + ".pb", "wb") as ff:
                ff.write(tp.SerializeToString())
            num_inputs = num_inputs + 1
        else:
            out_mp.graph.initializer.add().CopyFrom(i)

    for vi in list(out_mp.graph.input) + list(out_mp.graph.output) + list(out_mp.graph.value_info):
        dim = vi.type.tensor_type.shape.dim
        dim[len(dim) - 2].dim_param = "batch"

    for n in out_mp.graph.node:
        if n.op_type == "Scan":
            body = next(attr for attr in n.attribute if attr.name == "body")
            for vi in list(body.g.input) + list(body.g.output) + list(body.g.value_info):
                dim = vi.type.tensor_type.shape.dim
                dim[0].dim_param = "batch"

    onnx.save(out_mp, "test_Scan/model.onnx", "wb")


def GenSimpleScan():  # noqa: N802
    feature = C.sequence.input_variable((128,), np.float32)
    param = C.parameter(shape=(1,), dtype=np.float32)
    scan = C.layers.Recurrence(lambda h, x: x + h + param)(feature)
    model = C.sequence.reduce_sum(scan)
    data_feature = np.random.rand(1, 64, 128).astype(np.float32)
    data_output = np.asarray(model.eval(data_feature), dtype=np.float32)
    Save("test_SimpleScan", model, data_feature, data_output)


def GenGRU():  # noqa: N802
    feature = C.sequence.input_variable((64,), np.float32)
    gru_fw = C.layers.Recurrence(C.layers.GRU(128))(feature)
    gru_bw = C.layers.Recurrence(C.layers.GRU(128), go_backwards=True)(feature)
    model = C.splice(gru_fw, gru_bw, axis=0)
    data_feature = np.random.rand(1, 16, 64).astype(np.float32)
    data_output = np.asarray(model.eval(data_feature))
    Save("test_GRU", model, data_feature, data_output)


def GenRNN():  # noqa: N802
    feature = C.sequence.input_variable((64,), np.float32)
    model = C.optimized_rnnstack(
        feature,
        C.parameter(
            (
                C.InferredDimension,
                64,
            ),
            init=C.glorot_uniform(),
        ),
        128,
        2,
        True,
        "rnnReLU",
    )
    data_feature = np.random.rand(1, 16, 64).astype(np.float32)
    data_output = np.asarray(model.eval(data_feature))
    Save("test_RNN", model, data_feature, data_output)


if __name__ == "__main__":
    np.random.seed(0)
    GenSimple()
    GenSharedWeights()
    GenSimpleMNIST()
    GenMatMul_1k()
    GenLSTMx4(use_scan=True)
    GenLSTMx4(use_scan=False)
    GenSimpleScan()
    GenScan()
    GenGRU()
    GenRNN()



import onnx
from onnx import TensorProto, helper


def save_model(graph, file_name):
    model = helper.make_model(graph)
    onnx.checker.check_model(model)
    onnx.save(model, file_name)


graph = helper.make_graph(
    [  # nodes
        helper.make_node("Shape", ["SubgraphRoot"], ["shape1_out"], "shape1"),
        helper.make_node("Shape", ["SubgraphRoot"], ["shape2_out"], "shape2"),
        helper.make_node("Gather", ["shape1_out", "indices0"], ["gather0_out"], "gather0", axis=0),
        helper.make_node("Gather", ["shape1_out", "indices2"], ["gather2_out"], "gather2", axis=0),
        helper.make_node("Gather", ["shape2_out", "indices3"], ["gather3_out"], "gather3", axis=0),
        helper.make_node("Unsqueeze", ["gather0_out"], ["unsqueeze0_out"], "unsqueeze0", axes=[0]),
        helper.make_node("Unsqueeze", ["gather2_out"], ["unsqueeze2_out"], "unsqueeze2", axes=[0]),
        helper.make_node("Unsqueeze", ["gather3_out"], ["unsqueeze3_out"], "unsqueeze3", axes=[0]),
        helper.make_node(
            "Concat",
            ["unsqueeze0_out", "a1", "unsqueeze2_out", "unsqueeze3_out", "a4"],
            ["concat_out"],
            "concat",
            axis=0,
        ),
        helper.make_node("Reshape", ["SubgraphRoot", "concat_out"], ["Result"], "reshape"),
    ],
    "Reshape_Fusion",  # name
    [  # inputs
        helper.make_tensor_value_info("SubgraphRoot", TensorProto.FLOAT, ["unk_0", 256, "unk_2", "unk_3"]),
    ],
    [  # outputs
        helper.make_tensor_value_info("Result", TensorProto.FLOAT, ["unk_1", 128, "unk_2", "unk_3", "unk_4"]),
        helper.make_tensor_value_info("gather3_out", TensorProto.INT64, []),
    ],
    [  # initializers
        helper.make_tensor("a1", TensorProto.INT64, [1], [128]),
        helper.make_tensor("a4", TensorProto.INT64, [1], [-1]),
        helper.make_tensor("indices0", TensorProto.INT64, [], [0]),
        helper.make_tensor("indices2", TensorProto.INT64, [], [2]),
        helper.make_tensor("indices3", TensorProto.INT64, [], [3]),
    ],
)

save_model(graph, "reshape_fusion_internal_nodes_reused.onnx")


graph = helper.make_graph(
    [  # nodes
        helper.make_node("Shape", ["SubgraphRoot"], ["shape0_out"], "shape0"),
        helper.make_node("Shape", ["SubgraphRoot"], ["shape1_out"], "shape1"),
        helper.make_node("Gather", ["shape0_out", "indices0"], ["gather0_out"], "gather0", axis=0),
        helper.make_node("Gather", ["shape1_out", "indices1"], ["gather1_out"], "gather1", axis=0),
        helper.make_node("Unsqueeze", ["gather0_out"], ["unsqueeze0_out"], "unsqueeze0", axes=[0]),
        helper.make_node("Unsqueeze", ["gather1_out"], ["unsqueeze1_out"], "unsqueeze1", axes=[0]),
        helper.make_node(
            "Concat",
            ["unsqueeze0_out", "unsqueeze1_out", "a"],
            ["concat_out"],
            "concat",
            axis=0,
        ),
        helper.make_node("Reshape", ["SubgraphRoot", "concat_out"], ["Result"], "reshape"),
    ],
    "Reshape_Fusion",  # name
    [  # inputs
        helper.make_tensor_value_info("SubgraphRoot", TensorProto.FLOAT, [10, 20, 30]),
    ],
    [  # outputs
        helper.make_tensor_value_info("Result", TensorProto.FLOAT, [10, 20, "unk"]),
        helper.make_tensor_value_info("gather0_out", TensorProto.INT64, []),
    ],
    [  # initializers
        helper.make_tensor("a", TensorProto.INT64, [1], [-1]),
        helper.make_tensor("indices0", TensorProto.INT64, [], [0]),
        helper.make_tensor("indices1", TensorProto.INT64, [], [1]),
    ],
)

save_model(graph, "reshape_fusion_internal_node_is_graph_output.onnx")

graph = helper.make_graph(
    [  # nodes
        helper.make_node("Shape", ["SubgraphRoot"], ["shape2_out"], "shape2"),
        helper.make_node("Gather", ["shape2_out", "indices2"], ["gather2_out"], "gather2", axis=0),
        helper.make_node("Unsqueeze", ["gather2_out"], ["unsqueeze2_out"], "unsqueeze2", axes=[0]),
        helper.make_node("Concat", ["a", "unsqueeze2_out"], ["concat_out"], "concat", axis=0),
        helper.make_node("Reshape", ["SubgraphRoot", "concat_out"], ["Result"], "reshape"),
    ],
    "Reshape_Fusion",  # name
    [  # inputs
        helper.make_tensor_value_info("SubgraphRoot", TensorProto.FLOAT, [10, 20, 30]),
    ],
    [  # outputs
        helper.make_tensor_value_info("Result", TensorProto.FLOAT, ["unk_0", "unk_1", "unk_2"]),
    ],
    [  # initializers
        helper.make_tensor("a", TensorProto.INT64, [2], [1, 200]),
        helper.make_tensor("indices2", TensorProto.INT64, [], [1]),
    ],
)

save_model(graph, "reshape_fusion_multiple_values_in_initializer_tensor_1.onnx")

graph = helper.make_graph(
    [  # nodes
        helper.make_node("Shape", ["SubgraphRoot"], ["shape2_out"], "shape2"),
        helper.make_node("Gather", ["shape2_out", "indices2"], ["gather2_out"], "gather2", axis=0),
        helper.make_node("Unsqueeze", ["gather2_out"], ["unsqueeze2_out"], "unsqueeze2", axes=[0]),
        helper.make_node("Concat", ["a", "unsqueeze2_out"], ["concat_out"], "concat", axis=0),
        helper.make_node("Reshape", ["SubgraphRoot", "concat_out"], ["Result"], "reshape"),
    ],
    "Reshape_Fusion",  # name
    [  # inputs
        helper.make_tensor_value_info("SubgraphRoot", TensorProto.FLOAT, [10, 20, 30]),
    ],
    [  # outputs
        helper.make_tensor_value_info("Result", TensorProto.FLOAT, ["unk_0", "unk_1", "unk_2"]),
    ],
    [  # initializers
        helper.make_tensor("a", TensorProto.INT64, [2], [1, 200]),
        helper.make_tensor("indices2", TensorProto.INT64, [], [2]),
    ],
)

save_model(graph, "reshape_fusion_multiple_values_in_initializer_tensor_2.onnx")

graph = helper.make_graph(
    [  # nodes
        helper.make_node("Shape", ["AnotherInput"], ["shape2_out"], "shape2"),
        helper.make_node("Gather", ["shape2_out", "indices2"], ["gather2_out"], "gather2", axis=0),
        helper.make_node("Unsqueeze", ["gather2_out"], ["unsqueeze2_out"], "unsqueeze2", axes=[0]),
        helper.make_node("Concat", ["a", "unsqueeze2_out"], ["concat_out"], "concat", axis=0),
        helper.make_node("Reshape", ["SubgraphRoot", "concat_out"], ["Result"], "reshape"),
    ],
    "Reshape_Fusion",  # name
    [  # inputs
        helper.make_tensor_value_info("SubgraphRoot", TensorProto.FLOAT, [10, 20, 30]),
        helper.make_tensor_value_info(
            "AnotherInput",
            TensorProto.FLOAT,
            ["input_dim_0", "input_dim_1", "input_dim_2"],
        ),
    ],
    [  # outputs
        helper.make_tensor_value_info("Result", TensorProto.FLOAT, ["unk_0", "unk_1", "unk_2"]),
    ],
    [  # initializers
        helper.make_tensor("a", TensorProto.INT64, [2], [1, 200]),
        helper.make_tensor("indices2", TensorProto.INT64, [], [2]),
    ],
)

save_model(graph, "reshape_fusion_input_is_graph_input.onnx")

graph = helper.make_graph(
    [  # nodes
        helper.make_node("Concat", ["a"], ["concat_out"], "concat", axis=0),
        helper.make_node("Reshape", ["SubgraphRoot", "concat_out"], ["Result"], "reshape"),
    ],
    "Reshape_Fusion",  # name
    [  # inputs
        helper.make_tensor_value_info("SubgraphRoot", TensorProto.FLOAT, [2, 3, 4]),
        helper.make_tensor_value_info("a", TensorProto.INT64, [3]),
    ],
    [  # outputs
        helper.make_tensor_value_info("Result", TensorProto.FLOAT, ["unk_0", "unk_1", "unk_2"]),
    ],
    [  # initializers
        helper.make_tensor("a", TensorProto.INT64, [3], [1, 1, 2 * 3 * 4]),
    ],
)

save_model(graph, "reshape_fusion_overridable_initializer.onnx")

graph = helper.make_graph(
    [  # nodes
        helper.make_node("Shape", ["query"], ["shape0_out"], "shape0"),
        helper.make_node("Gather", ["shape0_out", "indices0"], ["gather0_out"], "gather0", axis=0),
        helper.make_node("Unsqueeze", ["gather0_out"], ["unsqueeze0_out"], "unsqueeze0", axes=[0]),
        helper.make_node("Concat", ["a", "unsqueeze0_out"], ["concat_out"], "concat", axis=0),
        helper.make_node("Reshape", ["doc_word_mask", "concat_out"], ["Result"], "reshape"),
    ],
    "Reshape_Fusion",  # name
    [  # inputs
        helper.make_tensor_value_info("query", TensorProto.FLOAT, [1, 50]),
        helper.make_tensor_value_info("doc_word_mask", TensorProto.FLOAT, [1, 200, 50]),
    ],
    [  # outputs
        helper.make_tensor_value_info("Result", TensorProto.FLOAT, [10, 20, "unk"]),
    ],
    [  # initializers
        helper.make_tensor("a", TensorProto.INT64, [1], [-1]),
        helper.make_tensor("indices0", TensorProto.INT64, [], [1]),
    ],
)

save_model(graph, "reshape_fusion_with_graph_inputs.onnx")

graph = helper.make_graph(
    [  # nodes
        helper.make_node("Shape", ["SubgraphRoot"], ["shape0_out"], "shape0"),
        helper.make_node("Shape", ["SubgraphRoot"], ["shape1_out"], "shape1"),
        helper.make_node("Gather", ["shape0_out", "indices0"], ["gather0_out"], "gather0", axis=0),
        helper.make_node("Gather", ["shape1_out", "indices1"], ["gather1_out"], "gather1", axis=0),
        helper.make_node("Unsqueeze", ["gather0_out"], ["unsqueeze0_out"], "unsqueeze0", axes=[0]),
        helper.make_node("Unsqueeze", ["gather1_out"], ["unsqueeze1_out"], "unsqueeze1", axes=[0]),
        helper.make_node("Shape", ["SubgraphRoot"], ["shape2_out"], "shape2"),
        helper.make_node(
            "Slice",
            ["shape2_out", "slice_starts", "slice_ends"],
            ["slice_out"],
            "slice1",
        ),
        helper.make_node(
            "Concat",
            ["unsqueeze0_out", "unsqueeze1_out", "slice_out"],
            ["concat_out"],
            "concat",
            axis=0,
        ),
        helper.make_node("Reshape", ["SubgraphRoot", "concat_out"], ["Result"], "reshape"),
    ],
    "Reshape_Fusion",  # name
    [  # inputs
        helper.make_tensor_value_info("SubgraphRoot", TensorProto.FLOAT, [10, 20, 30]),
    ],
    [  # outputs
        helper.make_tensor_value_info("Result", TensorProto.FLOAT, [10, 20, "unk"]),
    ],
    [  # initializers
        helper.make_tensor("indices0", TensorProto.INT64, [], [0]),
        helper.make_tensor("indices1", TensorProto.INT64, [], [1]),
        helper.make_tensor("slice_starts", TensorProto.INT64, [1], [2]),
        helper.make_tensor("slice_ends", TensorProto.INT64, [1], [3]),
    ],
)

save_model(graph, "reshape_fusion_concat_subgraph.onnx")

graph = helper.make_graph(
    [  # nodes
        helper.make_node("Shape", ["SubgraphRoot"], ["shape0_out"], "shape0"),
        helper.make_node("Shape", ["SubgraphRoot"], ["shape1_out"], "shape1"),
        helper.make_node("Gather", ["shape0_out", "indices0"], ["gather0_out"], "gather0", axis=0),
        helper.make_node("Gather", ["shape1_out", "indices1"], ["gather1_out"], "gather1", axis=0),
        helper.make_node("Unsqueeze", ["gather0_out"], ["unsqueeze0_out"], "unsqueeze0", axes=[0]),
        helper.make_node("Unsqueeze", ["gather1_out"], ["unsqueeze1_out"], "unsqueeze1", axes=[0]),
        helper.make_node("Shape", ["SubgraphRoot"], ["shape2_out"], "shape2"),
        helper.make_node("Slice", ["shape2_out"], ["slice_out"], "slice1", starts=[2], ends=[3]),
        helper.make_node(
            "Concat",
            ["unsqueeze0_out", "unsqueeze1_out", "slice_out"],
            ["concat_out"],
            "concat",
            axis=0,
        ),
        helper.make_node("Reshape", ["SubgraphRoot", "concat_out"], ["Result"], "reshape"),
    ],
    "Reshape_Fusion",  # name
    [  # inputs
        helper.make_tensor_value_info("SubgraphRoot", TensorProto.FLOAT, [10, 20, 30]),
    ],
    [  # outputs
        helper.make_tensor_value_info("Result", TensorProto.FLOAT, [10, 20, "unk"]),
    ],
    [  # initializers
        helper.make_tensor("indices0", TensorProto.INT64, [], [0]),
        helper.make_tensor("indices1", TensorProto.INT64, [], [1]),
    ],
)
# Save this model without checking
onnx.save(helper.make_model(graph), "reshape_fusion_with_slice1.onnx")

graph = helper.make_graph(
    [  # nodes
        helper.make_node("Shape", ["SubgraphRoot"], ["shape0_out"], "shape0"),
        helper.make_node("Shape", ["SubgraphRoot"], ["shape1_out"], "shape1"),
        helper.make_node("Gather", ["shape0_out", "indices0"], ["gather0_out"], "gather0", axis=0),
        helper.make_node("Gather", ["shape1_out", "indices1"], ["gather1_out"], "gather1", axis=0),
        helper.make_node("Unsqueeze", ["gather0_out"], ["unsqueeze0_out"], "unsqueeze0", axes=[0]),
        helper.make_node("Unsqueeze", ["gather1_out"], ["unsqueeze1_out"], "unsqueeze1", axes=[0]),
        helper.make_node("Shape", ["unsqueeze0_out"], ["dummy_out"], "dummy"),
        helper.make_node("Shape", ["SubgraphRoot"], ["shape2_out"], "shape2"),
        helper.make_node(
            "Slice",
            ["shape2_out", "slice_starts", "slice_ends"],
            ["slice_out"],
            "slice1",
        ),
        helper.make_node(
            "Concat",
            ["unsqueeze0_out", "unsqueeze1_out", "slice_out"],
            ["concat_out"],
            "concat",
            axis=0,
        ),
        helper.make_node("Reshape", ["SubgraphRoot", "concat_out"], ["Result"], "reshape"),
    ],
    "Reshape_Fusion",  # name
    [  # inputs
        helper.make_tensor_value_info("SubgraphRoot", TensorProto.FLOAT, [10, 20, 30]),
    ],
    [  # outputs
        helper.make_tensor_value_info("Result", TensorProto.FLOAT, [10, 20, "unk"]),
        helper.make_tensor_value_info("slice_out", TensorProto.INT64, [1]),
    ],
    [  # initializers
        helper.make_tensor("indices0", TensorProto.INT64, [], [0]),
        helper.make_tensor("indices1", TensorProto.INT64, [], [1]),
        helper.make_tensor("slice_starts", TensorProto.INT64, [1], [2]),
        helper.make_tensor("slice_ends", TensorProto.INT64, [1], [3]),
    ],
)

save_model(graph, "reshape_fusion_concat_subgraph_multiple_outputs.onnx")

graph = helper.make_graph(
    [  # nodes
        helper.make_node("Shape", ["SubgraphRoot"], ["shape0_out"], "shape0"),
        helper.make_node("Shape", ["SubgraphRoot"], ["shape1_out"], "shape1"),
        helper.make_node("Gather", ["shape0_out", "indices0"], ["gather0_out"], "gather0", axis=0),
        helper.make_node("Gather", ["shape1_out", "indices1"], ["gather1_out"], "gather1", axis=0),
        helper.make_node("Unsqueeze", ["gather0_out"], ["unsqueeze0_out"], "unsqueeze0", axes=[0]),
        helper.make_node("Unsqueeze", ["gather1_out"], ["unsqueeze1_out"], "unsqueeze1", axes=[0]),
        helper.make_node("Shape", ["SubgraphRoot"], ["shape2_out"], "shape2"),
        helper.make_node(
            "Slice",
            ["shape2_out", "slice_starts", "slice_ends"],
            ["slice_out"],
            "slice1",
        ),
        helper.make_node("Pad", ["slice_out", "pads"], ["pad0_out"], "pad0", mode="constant"),
        helper.make_node(
            "Concat",
            ["unsqueeze0_out", "unsqueeze1_out", "pad0_out"],
            ["concat_out"],
            "concat",
            axis=0,
        ),
        helper.make_node("Reshape", ["SubgraphRoot", "concat_out"], ["Result"], "reshape"),
    ],
    "Reshape_Fusion",  # name
    [  # inputs
        helper.make_tensor_value_info("SubgraphRoot", TensorProto.FLOAT, [10, 20, 30]),
    ],
    [helper.make_tensor_value_info("Result", TensorProto.FLOAT, [10, 20, "unk"])],  # outputs
    [  # initializers
        helper.make_tensor("indices0", TensorProto.INT64, [], [0]),
        helper.make_tensor("indices1", TensorProto.INT64, [], [1]),
        helper.make_tensor("pads", TensorProto.INT64, [2], [1, 0]),
        helper.make_tensor("slice_starts", TensorProto.INT64, [1], [2]),
        helper.make_tensor("slice_ends", TensorProto.INT64, [1], [3]),
    ],
)

save_model(graph, "reshape_fusion_concat_subgraph_not_triggered.onnx")

graph = helper.make_graph(
    [  # nodes
        helper.make_node("Shape", ["SubgraphRoot"], ["shape0_out"], "shape0"),
        helper.make_node("Shape", ["SubgraphRoot"], ["shape1_out"], "shape1"),
        helper.make_node("Gather", ["shape0_out", "indices0"], ["gather0_out"], "gather0", axis=0),
        helper.make_node("Gather", ["shape1_out", "indices1"], ["gather1_out"], "gather1", axis=0),
        helper.make_node("Unsqueeze", ["gather0_out"], ["unsqueeze0_out"], "unsqueeze0", axes=[0]),
        helper.make_node("Unsqueeze", ["gather1_out"], ["unsqueeze1_out"], "unsqueeze1", axes=[0]),
        helper.make_node("Shape", ["SubgraphRoot"], ["shape2_out"], "shape2"),
        helper.make_node(
            "Slice",
            ["shape2_out", "slice_starts", "slice_ends"],
            ["slice_out"],
            "slice1",
        ),
        helper.make_node("Squeeze", ["slice_out"], ["squeeze0_out"], "squeeze0", axes=[0]),
        helper.make_node("Div", ["squeeze0_out", "div_init"], ["div_out"], "div"),
        helper.make_node("Unsqueeze", ["div_out"], ["unsqueeze2_out"], "unsqueeze2", axes=[0]),
        helper.make_node(
            "Concat",
            ["unsqueeze0_out", "unsqueeze1_out", "unsqueeze2_out"],
            ["concat_out"],
            "concat",
            axis=0,
        ),
        helper.make_node("Reshape", ["SubgraphRoot", "concat_out"], ["Result"], "reshape"),
    ],
    "Reshape_Fusion",  # name
    [  # inputs
        helper.make_tensor_value_info("SubgraphRoot", TensorProto.FLOAT, [10, 20, 30]),
    ],
    [helper.make_tensor_value_info("Result", TensorProto.FLOAT, [10, 20, "unk"])],  # outputs
    [  # initializers
        helper.make_tensor("indices0", TensorProto.INT64, [], [0]),
        helper.make_tensor("indices1", TensorProto.INT64, [], [1]),
        helper.make_tensor("div_init", TensorProto.INT64, [], [1]),
        helper.make_tensor("slice_starts", TensorProto.INT64, [1], [2]),
        helper.make_tensor("slice_ends", TensorProto.INT64, [1], [3]),
    ],
)

save_model(graph, "reshape_fusion_concat_subgraph_div.onnx")

graph = helper.make_graph(
    [  # nodes
        helper.make_node("Shape", ["SubgraphRoot"], ["shape0_out"], "shape0"),
        helper.make_node("Shape", ["SubgraphRoot"], ["shape1_out"], "shape1"),
        helper.make_node("Gather", ["shape0_out", "indices0"], ["gather0_out"], "gather0", axis=0),
        helper.make_node("Gather", ["shape1_out", "indices1"], ["gather1_out"], "gather1", axis=0),
        helper.make_node("Unsqueeze", ["gather0_out"], ["unsqueeze0_out"], "unsqueeze0", axes=[0]),
        helper.make_node("Unsqueeze", ["gather1_out"], ["unsqueeze1_out"], "unsqueeze1", axes=[0]),
        helper.make_node("Shape", ["SubgraphRoot"], ["shape2_out"], "shape2"),
        helper.make_node(
            "Slice",
            ["shape2_out", "slice_starts_0", "slice_ends_0"],
            ["slice0_out"],
            "slice0",
        ),
        helper.make_node("Squeeze", ["slice0_out"], ["squeeze0_out"], "squeeze0", axes=[0]),
        helper.make_node("Shape", ["SubgraphRoot"], ["shape3_out"], "shape3"),
        helper.make_node(
            "Slice",
            ["shape3_out", "slice_starts_1", "slice_ends_1"],
            ["slice1_out"],
            "slice1",
        ),
        helper.make_node("Squeeze", ["slice1_out"], ["squeeze1_out"], "squeeze1", axes=[0]),
        helper.make_node("Mul", ["squeeze0_out", "squeeze1_out"], ["mul_out"], "mul"),
        helper.make_node("Unsqueeze", ["mul_out"], ["unsqueeze2_out"], "unsqueeze2", axes=[0]),
        helper.make_node(
            "Concat",
            ["unsqueeze0_out", "unsqueeze1_out", "unsqueeze2_out"],
            ["concat_out"],
            "concat",
            axis=0,
        ),
        helper.make_node("Reshape", ["SubgraphRoot", "concat_out"], ["Result"], "reshape"),
    ],
    "Reshape_Fusion",  # name
    [  # inputs
        helper.make_tensor_value_info("SubgraphRoot", TensorProto.FLOAT, [10, 20, 30]),
    ],
    [helper.make_tensor_value_info("Result", TensorProto.FLOAT, [10, 20, "unk"])],  # outputs
    [  # initializers
        helper.make_tensor("indices0", TensorProto.INT64, [], [0]),
        helper.make_tensor("indices1", TensorProto.INT64, [], [1]),
        helper.make_tensor("slice_starts_0", TensorProto.INT64, [1], [2]),
        helper.make_tensor("slice_ends_0", TensorProto.INT64, [1], [3]),
        helper.make_tensor("slice_starts_1", TensorProto.INT64, [1], [1]),
        helper.make_tensor("slice_ends_1", TensorProto.INT64, [1], [2]),
    ],
)

save_model(graph, "reshape_fusion_concat_subgraph_mul.onnx")

matmul_weights = [
    -0.04888916015625,
    0.0143280029296875,
    0.066650390625,
    -0.0343017578125,
    -0.0010356903076171875,
    -0.00048232078552246094,
    0.07470703125,
    -0.04736328125,
    0.01454925537109375,
    -0.0086669921875,
    -0.051971435546875,
    -0.0201568603515625,
    0.040435791015625,
    -0.019256591796875,
    0.0205078125,
    0.0111541748046875,
    0.0071868896484375,
    -0.0298309326171875,
    -0.0306549072265625,
    -0.0225372314453125,
    -0.04193115234375,
    0.07073974609375,
    -0.048065185546875,
    0.0198822021484375,
    -0.035552978515625,
    -0.022796630859375,
    0.03839111328125,
    0.007099151611328125,
    -0.0080108642578125,
    -0.0017957687377929688,
    0.0266265869140625,
    -0.028289794921875,
    0.0032901763916015625,
    0.0208740234375,
    -0.01529693603515625,
    -0.046600341796875,
    -0.034637451171875,
    0.011322021484375,
    -0.026458740234375,
    0.04656982421875,
    -0.0091705322265625,
    0.017913818359375,
    -0.019256591796875,
    -0.001216888427734375,
    -0.08245849609375,
    -0.023162841796875,
    -0.04132080078125,
    -0.03363037109375,
    0.0029315948486328125,
    0.03173828125,
    -0.004024505615234375,
    0.04534912109375,
    -0.0036163330078125,
    -0.03912353515625,
    -0.00800323486328125,
    0.058197021484375,
    0.05572509765625,
    0.01165771484375,
    0.06756591796875,
    0.05816650390625,
    -0.0654296875,
    -0.0241851806640625,
    0.0205535888671875,
    -0.031707763671875,
]

add_weight = [
    -0.23681640625,
    -0.16552734375,
    0.2191162109375,
    -0.1756591796875,
    -0.03460693359375,
    -0.05316162109375,
    -0.336181640625,
    -0.253662109375,
]

graph = helper.make_graph(
    [  # nodes
        helper.make_node("Add", ["Input", "Bias"], ["add0_out"], "add0"),
        helper.make_node("Shape", ["add0_out"], ["shape0_out"], "shape0"),
        helper.make_node("Gather", ["shape0_out", "indices0"], ["gather0_out"], "gather0", axis=0),
        helper.make_node("Unsqueeze", ["gather0_out"], ["unsqueeze0_out"], "unsqueeze0", axes=[0]),
        helper.make_node(
            "Concat",
            ["unsqueeze0_out", "dim_-1", "dim_2", "dim_4"],
            ["concat_out"],
            "concat",
            axis=0,
        ),
        helper.make_node("MatMul", ["add0_out", "matmul_weight"], ["matmul_out"], "matmul"),
        helper.make_node("Add", ["matmul_out", "add_weight"], ["add1_out"], "add1"),
        helper.make_node("Reshape", ["add1_out", "concat_out"], ["Result"], "reshape"),
    ],
    "Reshape_Fusion",  # name
    [  # inputs
        helper.make_tensor_value_info("Input", TensorProto.FLOAT, [1, 8]),
    ],
    [  # outputs
        helper.make_tensor_value_info("Result", TensorProto.FLOAT, [1, -1, 2, 4]),
    ],
    [  # initializers
        helper.make_tensor("Bias", TensorProto.FLOAT, [8], add_weight),
        helper.make_tensor("dim_-1", TensorProto.INT64, [1], [-1]),
        helper.make_tensor("dim_2", TensorProto.INT64, [1], [2]),
        helper.make_tensor("dim_4", TensorProto.INT64, [1], [4]),
        helper.make_tensor("indices0", TensorProto.INT64, [], [0]),
        helper.make_tensor("matmul_weight", TensorProto.FLOAT, [8, 8], matmul_weights),
        helper.make_tensor("add_weight", TensorProto.FLOAT, [8], add_weight),
    ],
)

save_model(graph, "reshape_fusion_distillbert.onnx")



import numpy as np
import onnx

from onnxruntime.quantization import CalibrationDataReader, QuantType, quantize
from onnxruntime.quantization.execution_providers.qnn import get_qnn_qdq_config

INPUT0_SHAPE = (1, 3, 8, 8)
INPUT0_NAME = "input_0"


def create_f32_model():
    input_0 = onnx.helper.make_tensor_value_info(INPUT0_NAME, onnx.TensorProto.FLOAT, INPUT0_SHAPE)
    output_0 = onnx.helper.make_tensor_value_info("output_0", onnx.TensorProto.FLOAT, None)
    weight_data = [
        [
            [[-1.5, -1.0, -0.5], [-0.2, 0.0, 0.2], [0.5, 1.0, 1.5]],  # range = 3.0, scale = 3.0/15, zp = 0
            [[-1.5, -1.0, -0.5], [-0.2, 0.0, 0.2], [0.5, 1.0, 1.5]],  # range = 3.0, scale = 3.0/15, zp = 0
            [[-1.5, -1.0, -0.5], [-0.2, 0.0, 0.2], [0.5, 1.0, 1.5]],  # range = 3.0, scale = 3.0/15, zp = 0
        ],
        [
            [[-1.0, -0.8, -0.6], [-0.1333, 0.0, -0.1333], [0.6, 0.8, 1.0]],  # range = 2.0, scale = 2.0/15, zp = -3
            [[-1.0, -0.8, -0.6], [-0.1333, 0.0, -0.1333], [0.6, 0.8, 1.0]],  # range = 2.0, scale = 2.0/15, zp = -3
            [[-1.0, -0.8, -0.6], [-0.1333, 0.0, -0.1333], [0.6, 0.8, 1.0]],  # range = 2.0, scale = 2.0/15, zp = -3
        ],
        [
            [[-1.5, -1.0, -0.5], [-0.2, 0.0, 0.2], [0.5, 1.0, 1.5]],  # range = 3.0, scale = 3.0/15, zp = 0
            [[-1.5, -1.0, -0.5], [-0.2, 0.0, 0.2], [0.5, 1.0, 1.5]],  # range = 3.0, scale = 3.0/15, zp = 0
            [[-1.5, -1.0, -0.5], [-0.2, 0.0, 0.2], [0.5, 1.0, 1.5]],  # range = 3.0, scale = 3.0/15, zp = 0
        ],
        [
            [[-1.0, -0.8, -0.6], [-0.1333, 0.0, -0.1333], [0.6, 0.8, 1.0]],  # range = 2.0, scale = 2.0/15, zp = -3
            [[-1.0, -0.8, -0.6], [-0.1333, 0.0, -0.1333], [0.6, 0.8, 1.0]],  # range = 2.0, scale = 2.0/15, zp = -3
            [[-1.0, -0.8, -0.6], [-0.1333, 0.0, -0.1333], [0.6, 0.8, 1.0]],  # range = 2.0, scale = 2.0/15, zp = -3
        ],
        [
            [[-1.5, -1.0, -0.5], [-0.2, 0.0, 0.2], [0.5, 1.0, 1.5]],  # range = 3.0, scale = 3.0/15, zp = 0
            [[-1.5, -1.0, -0.5], [-0.2, 0.0, 0.2], [0.5, 1.0, 1.5]],  # range = 3.0, scale = 3.0/15, zp = 0
            [[-1.5, -1.0, -0.5], [-0.2, 0.0, 0.2], [0.5, 1.0, 1.5]],  # range = 3.0, scale = 3.0/15, zp = 0
        ],
    ]
    weight = onnx.numpy_helper.from_array(np.array(weight_data, dtype=np.float32), "weight")
    bias_data = [-10.0, -8.0, 0.0, 8.0, 10.0]
    bias = onnx.numpy_helper.from_array(np.array(bias_data, dtype=np.float32), "bias")

    conv_node = onnx.helper.make_node("Conv", [INPUT0_NAME, "weight", "bias"], ["output_0"], name="Conv0")
    graph = onnx.helper.make_graph(
        [conv_node],
        "Convf32",
        [input_0],
        [output_0],
        initializer=[weight, bias],
    )
    opset_imports = [onnx.helper.make_opsetid("", 21)]
    model = onnx.helper.make_model(graph, opset_imports=opset_imports)
    model = onnx.shape_inference.infer_shapes(model)
    onnx.checker.check_model(model, True)

    return model


class DataReader(CalibrationDataReader):
    def __init__(self):
        self.enum_data = None
        self.data_list = []

        # Generate 10 random input values for calibration
        for _ in range(10):
            input_data = {INPUT0_NAME: np.random.random(INPUT0_SHAPE).astype(np.float32)}
            self.data_list.append(input_data)

        self.datasize = len(self.data_list)

    def get_next(self):
        if self.enum_data is None:
            self.enum_data = iter(self.data_list)
        return next(self.enum_data, None)

    def rewind(self):
        self.enum_data = None


def create_qdq_model(model_f32):
    # Use tensor quantization overrides to quantize Conv's weight input to 4 bits on axis 0.
    init_overrides = {"weight": [{"quant_type": QuantType.QInt4, "axis": 0, "symmetric": True}]}
    qnn_config = get_qnn_qdq_config(
        model_f32,
        DataReader(),
        init_overrides=init_overrides,
        activation_type=QuantType.QUInt16,
        weight_type=QuantType.QUInt8,
    )

    quantize(model_f32, "conv.int4_weights.qdq.onnx", qnn_config)


if __name__ == "__main__":
    model_f32 = create_f32_model()
    create_qdq_model(model_f32)



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
from pathlib import Path

import onnx
import onnx.helper as onnx_helper
import onnx.numpy_helper as onnx_numpy_helper
from onnx.onnx_pb import ModelProto

from .quant_utils import attribute_to_kwarg, find_by_name


def _clean_initializers_helper(graph, model):
    """Clean unused initializers from graph.

    Returns:
        A cleaned graph without unused initializers
        A list of tensor names, which are not produced by this graph and its subgraphes
    """
    requesting_tensor_names = set()
    requesting_tensor_names.update(input_name for node in graph.node for input_name in node.input if input_name)
    requesting_tensor_names.update(g_out.name for g_out in graph.output if g_out.name)

    new_nodes = []
    for node in graph.node:
        new_node = node
        graph_attrs = [
            attr
            for attr in node.attribute
            if attr.type == onnx.AttributeProto.GRAPH or attr.type == onnx.AttributeProto.GRAPHS
        ]
        if graph_attrs:
            kwargs = {}
            for attr in node.attribute:
                new_attribute = {}
                if attr.type == onnx.AttributeProto.GRAPH:
                    (
                        cleaned_sub_graph,
                        sub_requesting_tensor_names,
                    ) = _clean_initializers_helper(attr.g, model)
                    new_attribute = {attr.name: cleaned_sub_graph}
                    requesting_tensor_names.update(sub_requesting_tensor_names)
                elif attr.type == onnx.AttributeProto.GRAPHS:
                    cleaned_graphes = []
                    for subgraph in attr.graphs:
                        (
                            cleaned_sub_graph,
                            sub_requesting_tensor_names,
                        ) = _clean_initializers_helper(subgraph, model)
                        cleaned_graphes.append(cleaned_sub_graph)
                        requesting_tensor_names.update(sub_requesting_tensor_names)
                    new_attribute = {attr.name: cleaned_graphes}
                else:
                    new_attribute = attribute_to_kwarg(attr)
                kwargs.update(new_attribute)
            new_node = onnx_helper.make_node(node.op_type, node.input, node.output, name=node.name, **kwargs)
        new_nodes.append(new_node)

    graph.ClearField("node")
    graph.node.extend(new_nodes)

    requesting_tensor_names.difference_update(output for node in graph.node for output in node.output)

    unused_initializer = []
    for initializer in graph.initializer:
        if initializer.name in requesting_tensor_names:
            requesting_tensor_names.remove(initializer.name)
        else:
            # mark it to remove, remove here directly will cause mis-behavier
            unused_initializer.append(initializer)

    name_to_input = {input.name: input for input in graph.input}
    for initializer in unused_initializer:
        graph.initializer.remove(initializer)
        if initializer.name in name_to_input:
            try:
                graph.input.remove(name_to_input[initializer.name])
            except StopIteration:
                if model.ir_version < 4:
                    print(f"Warning: invalid weight name {initializer.name} found in the graph (not a graph input)")

    requesting_tensor_names.difference_update(input.name for input in graph.input)

    return graph, requesting_tensor_names


class ONNXModel:
    def __init__(self, model: ModelProto):
        self.model = model

    def nodes(self):
        return self.model.graph.node

    def initializer(self):
        return self.model.graph.initializer

    def initializer_extend(self, inits):
        if len(inits) == 0:
            raise ValueError("Can add an empty list.")
        for init in self.initializer():
            self._check_init(init, "gain")
        for init in inits:
            self._check_init(init)
            self.model.graph.initializer.append(init)

    def graph(self):
        return self.model.graph

    def ir_version(self):
        return self.model.ir_version

    def opset_import(self):
        return self.model.opset_import

    def set_opset_import(self, domain, version):
        for opset in self.model.opset_import:
            if opset.domain == domain:
                opset.version = version
                return

        self.model.opset_import.extend([onnx_helper.make_opsetid(domain, version)])

    def remove_node(self, node):
        if node in self.model.graph.node:
            self.model.graph.node.remove(node)

    def remove_nodes(self, nodes_to_remove):
        for node in nodes_to_remove:
            self.remove_node(node)

    def add_node(self, node):
        self.model.graph.node.extend([self._check_node(node)])

    def add_nodes(self, nodes_to_add):
        for node in nodes_to_add:
            self.add_node(node)

    def add_initializer(self, tensor):
        if find_by_name(tensor.name, self.model.graph.initializer) is None:
            self._check_init(tensor)
            self.model.graph.initializer.extend([tensor])

    def get_initializer(self, name):
        for tensor in self.model.graph.initializer:
            if tensor.name == name:
                return tensor
        return None

    def find_graph_input(self, input_name):
        for input in self.model.graph.input:
            if input.name == input_name:
                return input
        return None

    def find_graph_output(self, output_name):
        for output in self.model.graph.output:
            if output.name == output_name:
                return output
        return None

    def get_tensor_type(self, tensor_name: str):
        tensor_type_map = {obj.name: obj.type for obj in self.model.graph.value_info}

        if tensor_name in tensor_type_map:
            return tensor_type_map[tensor_name].tensor_type

        g_input = self.find_graph_input(tensor_name)
        if g_input:
            return g_input.type.tensor_type

        g_output = self.find_graph_output(tensor_name)
        if g_output:
            return g_output.type.tensor_type

        return None

    def get_constant_value(self, output_name):
        for node in self.model.graph.node:
            if node.op_type == "Constant":
                if node.output[0] == output_name:
                    for attr in node.attribute:
                        if attr.name == "value":
                            return onnx_numpy_helper.to_array(attr.t)

        # Fallback to initializer since constant folding may have been applied.
        initializer = self.get_initializer(output_name)
        if initializer is not None:
            return onnx_numpy_helper.to_array(initializer)

        return None

    def get_initializer_name_set(self):
        return {initializer.name for initializer in self.model.graph.initializer}

    def remove_initializer(self, tensor):
        if tensor in self.model.graph.initializer:
            self.model.graph.initializer.remove(tensor)
            for input in self.model.graph.input:
                if input.name == tensor.name:
                    self.model.graph.input.remove(input)
                    break

    def remove_initializers(self, init_to_remove):
        for initializer in init_to_remove:
            self.remove_initializer(initializer)

    def get_non_initializer_inputs(self):
        initializer_names = self.get_initializer_name_set()
        non_initializer_inputs = set()
        for input in self.model.graph.input:
            if input.name not in initializer_names:
                non_initializer_inputs.add(input.name)
        return non_initializer_inputs

    def input_name_to_nodes(self):
        input_name_to_nodes = {}
        for node in self.model.graph.node:
            for input_name in node.input:
                if input_name:  # Could be empty when it is optional
                    if input_name not in input_name_to_nodes:
                        input_name_to_nodes[input_name] = [node]
                    else:
                        input_name_to_nodes[input_name].append(node)
        return input_name_to_nodes

    def output_name_to_node(self):
        output_name_to_node = {}
        for node in self.model.graph.node:
            for output_name in node.output:
                if output_name:  # Could be empty when it is optional
                    output_name_to_node[output_name] = node
        return output_name_to_node

    def get_children(self, node, input_name_to_nodes=None):
        if input_name_to_nodes is None:
            input_name_to_nodes = self.input_name_to_nodes()

        children = []
        for output in node.output:
            if output in input_name_to_nodes:
                for node in input_name_to_nodes[output]:
                    children.append(node)  # noqa: PERF402
        return children

    def get_parents(self, node, output_name_to_node=None):
        if output_name_to_node is None:
            output_name_to_node = self.output_name_to_node()

        parents = []
        for input in node.input:
            if input in output_name_to_node:
                parents.append(output_name_to_node[input])
        return parents

    def get_parent(self, node, idx, output_name_to_node=None):
        if output_name_to_node is None:
            output_name_to_node = self.output_name_to_node()

        if len(node.input) <= idx:
            return None

        input = node.input[idx]
        if input not in output_name_to_node:
            return None

        return output_name_to_node[input]

    def find_node_by_name(self, node_name, new_nodes_list, graph):
        """Find out if a node exists in a graph or a node is in the
        new set of nodes created during quantization.

        Returns:
            The node found or None.
        """
        graph_nodes_list = list(graph.node)  # deep copy
        graph_nodes_list.extend(new_nodes_list)
        node = find_by_name(node_name, graph_nodes_list)
        return node

    def get_largest_node_name_suffix(self, node_name_prefix):
        """
        Gets the largest node name (int) suffix for all node names that begin with `node_name_prefix`.
        Example: for nodes my_prefix_0 and my_prefix_3, this method returns 3.
        """
        suffix = -1

        for node in self.model.graph.node:
            if node.name and node.name.startswith(node_name_prefix):
                try:
                    index = int(node.name[len(node_name_prefix) :])
                    suffix = max(index, suffix)
                except ValueError:
                    continue

        return suffix

    def get_largest_initializer_name_suffix(self, initializer_name_prefix):
        """
        Gets the largest initializer name integer suffix for all initializer names that begin
        with `initializer_name_prefix`. This can be used to create unique initializer names.

        Example: for initializer names 'my_weight_0' and 'my_weight_3', this method returns 3 if
                 `initializer_name_prefix` is 'my_weight_'.
        """
        suffix = -1

        for initializer in self.model.graph.initializer:
            if initializer.name.startswith(initializer_name_prefix):
                try:
                    index = int(initializer.name[len(initializer_name_prefix) :])
                    suffix = max(index, suffix)
                except ValueError:
                    continue

        return suffix

    def find_nodes_by_initializer(self, graph, initializer):
        """
        Find all nodes with given initializer as an input.
        """
        nodes = []
        for node in graph.node:
            for node_input in node.input:
                if node_input == initializer.name:
                    nodes.append(node)
        return nodes

    @staticmethod
    def __get_initializer(name, graph_path):
        for gid in range(len(graph_path) - 1, -1, -1):
            graph = graph_path[gid]
            for tensor in graph.initializer:
                if tensor.name == name:
                    return tensor, graph
        return None, None

    @staticmethod
    def __replace_gemm_with_matmul(graph_path):
        new_nodes = []
        graph = graph_path[-1]
        for node in graph.node:
            graph_attrs = [attr for attr in node.attribute if attr.type == 5 or attr.type == 10]
            if len(graph_attrs):
                kwargs = {}
                for attr in node.attribute:
                    if attr.type == 5:
                        graph_path.append(attr.g)
                        kv = {attr.name: ONNXModel.__replace_gemm_with_matmul(graph_path)}
                    elif attr.type == 10:
                        value = []
                        for subgraph in attr.graphs:
                            graph_path.append(subgraph)
                            value.extend([ONNXModel.__replace_gemm_with_matmul(graph_path)])
                        kv = {attr.name: value}
                    else:
                        kv = attribute_to_kwarg(attr)
                    kwargs.update(kv)
                node = onnx_helper.make_node(  # noqa: PLW2901
                    node.op_type, node.input, node.output, name=node.name, **kwargs
                )

            if node.op_type == "Gemm":
                alpha = 1.0
                beta = 1.0
                transA = 0  # noqa: N806
                transB = 0  # noqa: N806
                for attr in node.attribute:
                    if attr.name == "alpha":
                        alpha = onnx_helper.get_attribute_value(attr)
                    elif attr.name == "beta":
                        beta = onnx_helper.get_attribute_value(attr)
                    elif attr.name == "transA":
                        transA = onnx_helper.get_attribute_value(attr)  # noqa: N806
                    elif attr.name == "transB":
                        transB = onnx_helper.get_attribute_value(attr)  # noqa: N806
                if alpha == 1.0 and beta == 1.0 and transA == 0:
                    inputB = node.input[1]  # noqa: N806
                    if transB == 1:
                        B, Bs_graph = ONNXModel.__get_initializer(node.input[1], graph_path)  # noqa: N806
                        if B:
                            # assume B is not used by any other node
                            B_array = onnx_numpy_helper.to_array(B)  # noqa: N806
                            B_trans = onnx_numpy_helper.from_array(B_array.T)  # noqa: N806
                            B_trans.name = B.name
                            Bs_graph.initializer.remove(B)
                            for input in Bs_graph.input:
                                if input.name == inputB:
                                    Bs_graph.input.remove(input)
                                    break
                            Bs_graph.initializer.extend([B_trans])
                        else:
                            inputB += "_Transposed"  # noqa: N806
                            transpose_node = onnx_helper.make_node(
                                "Transpose",
                                inputs=[node.input[1]],
                                outputs=[inputB],
                                name=node.name + "_Transpose" if node.name else "",
                            )
                            new_nodes.append(transpose_node)

                    matmul_node = onnx_helper.make_node(
                        "MatMul",
                        inputs=[node.input[0], inputB],
                        outputs=[node.output[0] + ("_MatMul" if len(node.input) > 2 else "")],
                        name=node.name + "_MatMul" if node.name else "",
                    )
                    new_nodes.append(matmul_node)

                    if len(node.input) > 2:
                        add_node = onnx_helper.make_node(
                            "Add",
                            inputs=[node.output[0] + "_MatMul", node.input[2]],
                            outputs=node.output,
                            name=node.name + "_Add" if node.name else "",
                        )
                        new_nodes.append(add_node)

                # unsupported
                else:
                    new_nodes.append(node)

            # not GEMM
            else:
                new_nodes.append(node)

        graph.ClearField("node")
        graph.node.extend(new_nodes)
        graph_path.pop()
        return graph

    def replace_gemm_with_matmul(self):
        graph_path = [self.graph()]
        ONNXModel.__replace_gemm_with_matmul(graph_path)

    def save_model_to_file(self, output_path, use_external_data_format=False):
        """
        Save model to external data, which is needed for model size > 2GB
        """
        self.topological_sort()
        if use_external_data_format:
            onnx.external_data_helper.convert_model_to_external_data(
                self.model,
                all_tensors_to_one_file=True,
                location=Path(output_path).name + ".data",
                convert_attribute=True,
            )
        for init in self.model.graph.initializer:
            self._check_init(init, "end")
        onnx.save_model(self.model, output_path)

    @staticmethod
    def replace_node_input(node, old_input_name, new_input_name):
        assert isinstance(old_input_name, str) and isinstance(new_input_name, str)
        for j in range(len(node.input)):
            if node.input[j] == old_input_name:
                node.input[j] = new_input_name

    def replace_input_of_all_nodes(self, old_input_name, new_input_name):
        for node in self.model.graph.node:
            ONNXModel.replace_node_input(node, old_input_name, new_input_name)

    def replace_input_of_nodes(self, old_input_name, new_input_name, node_names_set):
        for node in self.model.graph.node:
            if node.name in node_names_set:
                ONNXModel.replace_node_input(node, old_input_name, new_input_name)

    @staticmethod
    def replace_node_output(node, old_output_name, new_output_name):
        assert isinstance(old_output_name, str) and isinstance(new_output_name, str)
        for j in range(len(node.output)):
            if node.output[j] == old_output_name:
                node.output[j] = new_output_name

    def replace_output_of_all_nodes(self, old_output_name, new_output_name):
        for node in self.model.graph.node:
            ONNXModel.replace_node_output(node, old_output_name, new_output_name)

    def replace_output_of_nodes(self, old_output_name, new_output_name, node_names_set):
        for node in self.model.graph.node:
            if node.name in node_names_set:
                ONNXModel.replace_node_output(node, old_output_name, new_output_name)

    def remove_unused_constant(self):
        input_name_to_nodes = self.input_name_to_nodes()

        # remove unused constant
        unused_nodes = []
        nodes = self.nodes()
        for node in nodes:
            if (
                node.op_type == "Constant"
                and not self.is_graph_output(node.output[0])
                and node.output[0] not in input_name_to_nodes
            ):
                unused_nodes.append(node)

        self.remove_nodes(unused_nodes)

        ununsed_weights = []
        for w in self.initializer():
            if w.name not in input_name_to_nodes and not self.is_graph_output(w.name):
                ununsed_weights.append(w)
                # Remove from graph.input
                for graph_input in self.graph().input:
                    if graph_input.name == w.name:
                        self.graph().input.remove(graph_input)

        self.remove_initializers(ununsed_weights)

    def is_graph_output(self, output_name):
        return any(output.name == output_name for output in self.model.graph.output)

    def is_graph_input(self, tensor_name: str) -> bool:
        return any(input.name == tensor_name for input in self.model.graph.input)

    # TODO:use OnnxModel.graph_topological_sort(self.model.graph) from transformers.onnx_model
    # Currently it breaks Openvino/Linux training gpu pipeline so hold off for 1.8 release
    def topological_sort(self):
        deps_count = [0] * len(self.nodes())  # dependency count of each node
        deps_to_nodes = {}  # input to node indice
        sorted_nodes = []  # initialize sorted_nodes
        for node_idx, node in enumerate(self.nodes()):
            # CANNOT use len(node.input) directly because input can be optional
            deps_count[node_idx] = sum(1 for _ in node.input if _)
            if deps_count[node_idx] == 0:  # Constant doesn't depend on any inputs
                sorted_nodes.append(self.nodes()[node_idx])
                continue

            for input_name in node.input:
                if not input_name:
                    continue
                if input_name not in deps_to_nodes:
                    deps_to_nodes[input_name] = [node_idx]
                else:
                    deps_to_nodes[input_name].append(node_idx)

        initializer_names = [init.name for init in self.initializer()]
        graph_input_names = [input.name for input in self.model.graph.input]
        input_names = initializer_names + graph_input_names
        input_names.sort()
        prev_input_name = None
        for input_name in input_names:
            if prev_input_name == input_name:
                continue

            prev_input_name = input_name
            if input_name in deps_to_nodes:
                for node_idx in deps_to_nodes[input_name]:
                    deps_count[node_idx] = deps_count[node_idx] - 1
                    if deps_count[node_idx] == 0:
                        sorted_nodes.append(self.nodes()[node_idx])

        start = 0
        end = len(sorted_nodes)

        while start < end:
            for output in sorted_nodes[start].output:
                if output in deps_to_nodes:
                    for node_idx in deps_to_nodes[output]:
                        deps_count[node_idx] = deps_count[node_idx] - 1
                        if deps_count[node_idx] == 0:
                            sorted_nodes.append(self.nodes()[node_idx])
                            end = end + 1
            start = start + 1

        assert end == len(self.graph().node), "Graph is not a DAG"
        self.graph().ClearField("node")
        self.graph().node.extend(sorted_nodes)

    def clean_initializers(self):
        return _clean_initializers_helper(self.graph(), self.model)

    def _check_init(self, init, test=None):
        if init.data_type == onnx.TensorProto.FLOAT8E4M3FN:
            if init.HasField("raw_data"):
                b = list(init.raw_data)
                if any((i & 127) == 127 for i in b):
                    raise ValueError(f"Initializer {init.name!r} has nan.")
        return init

    def _check_node(self, node):
        """
        A quantization to float 8 does not use quantized bias but float 16 bias.
        This function checks that DequantizeLinear is not used to
        dequantize from float 16.
        """
        if node.op_type == "DequantizeLinear":
            zero_point = node.input[2]
            init = self.get_initializer(zero_point)
            dtype = init.data_type
            if dtype in {
                onnx.TensorProto.FLOAT16,
                onnx.TensorProto.FLOAT,
                onnx.TensorProto.DOUBLE,
                onnx.TensorProto.BFLOAT16,
            }:
                raise RuntimeError(f"Unsupported DequantizeLinear operator, dequantization from {dtype}.")
        return node



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

"""
PyTorch's _efficient_attention_forward/_efficient_attention_backward APIs is keep changing. Current implementation
is tested well on version 2.3.0.dev20240221+cu118, and should be run well since official version 2.3.0. If may fail to
run is you are using PyTorch with older versions.

This file is more like an example of how to add a new graph optimizer. Ideally user can add graph optimizer according
to the specific model they are using on their own instead of putting every possible graph optimizer here.

PyTorch also has API for flash attention (currently doesn't support random attention mask or Dropout), we can add
support if we want to try in the future.
"""

from onnx import GraphProto, NodeProto, TensorProto, helper

from ..graph_optimizer_registry import register_graph_optimizer
from .utils import GraphMatcher, check_attribute_value, make_constant_node, update_graph


def _make_efficient_attention_nodes(
    idx: int,
    q: str,
    k: str,
    v: str,
    y: str,
    dy: str,
    dq: str,
    dk: str,
    dv: str,
    bias: str,
    expand_bias: bool,
    scale: float,
    dropout_ratio: float,
    causal: bool,
):
    nodes_to_add = []
    scale_node = make_constant_node("scale_" + str(idx), TensorProto.FLOAT, [], [scale])
    dropout_ratio_node = make_constant_node("dropout_ratio_" + str(idx), TensorProto.FLOAT, [], [dropout_ratio])
    causal_node = make_constant_node("causal_" + str(idx), TensorProto.INT64, [], [1 if causal else 0])
    one_node = make_constant_node("one_" + str(idx), TensorProto.INT64, [], [1])
    zero_node = make_constant_node("zero_" + str(idx), TensorProto.INT64, [], [0])
    logsumexp = helper.make_tensor_value_info("logsumexp" + str(idx), TensorProto.FLOAT, [])
    seed = helper.make_tensor_value_info("seed" + str(idx), TensorProto.INT64, [])
    offset = helper.make_tensor_value_info("offset" + str(idx), TensorProto.INT64, [])
    msb_q = helper.make_tensor_value_info("msb_q_" + str(idx), TensorProto.INT64, [])
    msb_k = helper.make_tensor_value_info("msb_k_" + str(idx), TensorProto.INT64, [])
    new_value_infos = [logsumexp, seed, offset, msb_q, msb_k]
    if expand_bias:
        shape_0 = helper.make_node("Shape", [q], ["shape_0_" + str(idx)], start=0, end=1)
        shape_1 = helper.make_node("Shape", [q], ["shape_1_" + str(idx)], start=2, end=3)
        shape_2 = helper.make_node("Shape", [q], ["shape_2_" + str(idx)], start=1, end=2)
        shape_3 = helper.make_node("Shape", [k], ["shape_3_" + str(idx)], start=1, end=2)
        concat = helper.make_node(
            "Concat",
            [shape_0.output[0], shape_1.output[0], shape_2.output[0], shape_3.output[0]],
            ["concated_shape_" + str(idx)],
            axis=0,
        )
        expand = helper.make_node("Expand", [bias, concat.output[0]], ["expanded_bias_" + str(idx)])
        nodes_to_add.extend([shape_0, shape_1, shape_2, shape_3, concat, expand])
        bias = expand.output[0]
    fwd_node = helper.make_node(
        "ATen",
        [
            q,
            k,
            v,
            bias,
            "",
            "",
            "",
            "",
            dropout_ratio_node.output[0],
            causal_node.output[0],
            one_node.output[0],
            scale_node.output[0],
            "",
            "",
        ],
        [y, logsumexp.name, seed.name, offset.name, msb_q.name, msb_k.name],
        "efficient_attention_forward_" + str(idx),
        None,
        "org.pytorch.aten",
        operator="_efficient_attention_forward",
        cpu_input_args=[4, 5, 12, 13],
        cpu_output_args=[2, 3, 4, 5],
    )
    bwd_node = helper.make_node(
        "ATen",
        [
            dy,
            q,
            k,
            v,
            bias,
            y,
            "",
            "",
            msb_q.name,
            msb_k.name,
            logsumexp.name,
            dropout_ratio_node.output[0],
            seed.name,
            offset.name,
            causal_node.output[0],
            zero_node.output[0],
            scale_node.output[0],
            "",
        ],
        [dq, dk, dv, ""],
        "efficient_attention_backward_" + str(idx),
        None,
        "org.pytorch.aten",
        operator="_efficient_attention_backward",
        cpu_input_args=[6, 7, 12, 13],
    )
    nodes_to_add.extend([scale_node, dropout_ratio_node, causal_node, one_node, zero_node, fwd_node, bwd_node])
    return nodes_to_add, new_value_infos


# Without causal mask, with Dropout. For example, BERT model in HuggingFace.
_PATTERN_0: list[tuple[str, bool, list[tuple[int, int, int]]]] = [
    ("MatMul", False, []),  # 0
    ("Transpose", True, [(0, 0, 0)]),  # 1
    ("Transpose", True, [(0, 0, 1)]),  # 2
    ("Div", False, [(0, 0, 0)]),  # 3
    ("Add", False, [(3, 0, 0)]),  # 4
    ("Softmax", False, [(4, 0, 0)]),  # 5
    ("Dropout", False, [(5, 0, 0)]),  # 6
    ("MatMul", False, [(6, 0, 0)]),  # 7
    ("Transpose", True, [(7, 0, 1)]),  # 8
    ("Transpose", False, [(7, 0, 0)]),  # 9
    ("FusedMatMul", False, [(8, 0, 1)]),  # 10
    ("DropoutGrad", False, [(10, 0, 0), (6, 1, 1)]),  # 11
    ("SoftmaxGrad_13", False, [(11, 0, 0), (5, 0, 1)]),  # 12
    ("Identity", False, [(12, 0, 0)]),  # 13
    ("Div", False, [(13, 0, 0)]),  # 14
    ("Identity", False, [(14, 0, 0)]),  # 15
    ("FusedMatMul", False, [(2, 0, 1), (15, 0, 0)]),  # 16
    ("FusedMatMul", False, [(1, 0, 0), (15, 0, 1)]),  # 17
    ("FusedMatMul", False, [(6, 0, 0)]),  # 18
    ("Transpose", True, [(18, 0, 1)]),  # 19
    ("Transpose", False, [(16, 0, 0)]),  # 20
    ("Transpose", False, [(17, 0, 0)]),  # 21
    ("Transpose", False, [(18, 0, 0)]),  # 22
]


def _optimize_for_pattern_0(matcher: GraphMatcher, idx: int, nodes: list[NodeProto]):
    # Check forward only as the backward is expected to be consistent if it's built correctly.
    scale_value = matcher.get_constant_value(nodes[3].input[1])
    ratio_value = matcher.get_constant_value(nodes[6].input[1])
    if not (
        check_attribute_value(nodes[1], "perm", [0, 2, 1, 3])
        and check_attribute_value(nodes[2], "perm", [0, 2, 3, 1])
        and scale_value is not None
        and ratio_value is not None
        and check_attribute_value(nodes[8], "perm", [0, 2, 1, 3])
        and check_attribute_value(nodes[9], "perm", [0, 2, 1, 3])
    ):
        return [], [], []

    _, add_input_shape_0 = matcher.get_type_and_shape(nodes[4].input[0])
    _, add_input_shape_1 = matcher.get_type_and_shape(nodes[4].input[1])
    nodes_to_add, new_value_infos = _make_efficient_attention_nodes(
        idx,
        nodes[1].input[0],
        nodes[2].input[0],
        nodes[8].input[0],
        nodes[9].output[0],
        nodes[19].input[0],
        nodes[20].output[0],
        nodes[21].output[0],
        nodes[22].output[0],
        nodes[4].input[1],
        add_input_shape_0 != add_input_shape_1,
        1 / float(scale_value[0] if isinstance(scale_value, list) else scale_value),
        ratio_value,
        False,
    )
    return nodes, nodes_to_add, new_value_infos


# Without causal mask, without Dropout. For example, BERT model and disabling attention dropout in HuggingFace.
_PATTERN_1: list[tuple[str, bool, list[tuple[int, int, int]]]] = [
    ("MatMul", False, []),  # 0
    ("Transpose", True, [(0, 0, 0)]),  # 1
    ("Transpose", True, [(0, 0, 1)]),  # 2
    ("Div", False, [(0, 0, 0)]),  # 3
    ("Add", False, [(3, 0, 0)]),  # 4
    ("Softmax", False, [(4, 0, 0)]),  # 5
    ("MatMul", False, [(5, 0, 0)]),  # 6
    ("Transpose", True, [(6, 0, 1)]),  # 7
    ("Transpose", False, [(6, 0, 0)]),  # 8
    ("FusedMatMul", False, [(7, 0, 1)]),  # 9
    ("SoftmaxGrad_13", False, [(9, 0, 0), (5, 0, 1)]),  # 10
    ("Identity", False, [(10, 0, 0)]),  # 11
    ("Div", False, [(11, 0, 0)]),  # 12
    ("Identity", False, [(12, 0, 0)]),  # 13
    ("FusedMatMul", False, [(2, 0, 1), (13, 0, 0)]),  # 14
    ("FusedMatMul", False, [(1, 0, 0), (13, 0, 1)]),  # 15
    ("FusedMatMul", False, [(5, 0, 0)]),  # 16
    ("Transpose", True, [(16, 0, 1)]),  # 17
    ("Transpose", False, [(14, 0, 0)]),  # 18
    ("Transpose", False, [(15, 0, 0)]),  # 19
    ("Transpose", False, [(16, 0, 0)]),  # 20
]


def _optimize_for_pattern_1(matcher: GraphMatcher, idx: int, nodes: list[NodeProto]):
    # Check forward only as the backward is expected to be consistent if it's built correctly.
    scale_value = matcher.get_constant_value(nodes[3].input[1])
    if not (
        check_attribute_value(nodes[1], "perm", [0, 2, 1, 3])
        and check_attribute_value(nodes[2], "perm", [0, 2, 3, 1])
        and scale_value is not None
        and check_attribute_value(nodes[7], "perm", [0, 2, 1, 3])
        and check_attribute_value(nodes[8], "perm", [0, 2, 1, 3])
    ):
        return [], [], []

    _, add_input_shape_0 = matcher.get_type_and_shape(nodes[4].input[0])
    _, add_input_shape_1 = matcher.get_type_and_shape(nodes[4].input[1])
    nodes_to_add, new_value_infos = _make_efficient_attention_nodes(
        idx,
        nodes[1].input[0],
        nodes[2].input[0],
        nodes[7].input[0],
        nodes[8].output[0],
        nodes[17].input[0],
        nodes[18].output[0],
        nodes[19].output[0],
        nodes[20].output[0],
        nodes[4].input[1],
        add_input_shape_0 != add_input_shape_1,
        1 / float(scale_value[0] if isinstance(scale_value, list) else scale_value),
        0.0,
        False,
    )
    return nodes, nodes_to_add, new_value_infos


_PATTERNS = [
    (_PATTERN_0, _optimize_for_pattern_0),
    (_PATTERN_1, _optimize_for_pattern_1),
]


@register_graph_optimizer(devices="cuda")
def optimize_graph_for_aten_efficient_attention(graph: GraphProto):
    nodes_to_remove = []
    nodes_to_add = []
    new_value_infos = []
    matcher = GraphMatcher(graph)
    idx = 0
    for pattern_tuple in _PATTERNS:
        for nodes in matcher.match_pattern(pattern_tuple[0]):
            remove_nodes, add_nodes, add_value_infos = pattern_tuple[1](matcher, idx, nodes)
            if len(add_nodes) > 0:
                nodes_to_remove.extend(remove_nodes)
                nodes_to_add.extend(add_nodes)
                new_value_infos.extend(add_value_infos)
                idx += 1
    update_graph(graph, nodes_to_remove, nodes_to_add, new_value_infos)



#!/usr/bin/env python
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""
Tests quantization of ConvTranspose operator.
"""

import unittest

import numpy as np
import onnx
from onnx import TensorProto, helper
from op_test_utils import TestDataFeeds, check_model_correctness, check_op_type_count, check_qtype_by_node_type

from onnxruntime.quantization import QuantFormat, QuantType, quantize_static


class TestOpConvTranspose(unittest.TestCase):
    """
    Class with test_* methods that test quantization of the ConvTranspose operator.
    """

    def input_feeds(self, num_test_inputs, name2shape, dtype):
        """
        Returns a data reader of input test data.

        :param num_test_inputs: The number of testing inputs to generate.
        :param name2shape: A dictionary mapping the name of an input to its shape.

        :return: A data reader.
        """
        input_data_list = []
        for _ in range(num_test_inputs):
            inputs = {}
            for name, shape in name2shape.items():
                inputs.update({name: np.random.randint(-1, 2, shape).astype(dtype)})
            input_data_list.extend([inputs])
        data_reader = TestDataFeeds(input_data_list)
        return data_reader

    def construct_model(self, output_model_path, onnx_type=TensorProto.FLOAT, opset=13, ir_version=7):
        """
        Constructs an ONNX model containing a single ConvTranspose node, and saves
        the model to the specified output path.

        :param output_model_path: The output filepath in which to save the model.
        """

        input_tensor = helper.make_tensor_value_info("input", onnx_type, [1, 1, 7, 7])
        output_tensor = helper.make_tensor_value_info("output", onnx_type, [1, 1, 8, 8])
        ini_w = helper.make_tensor("weight", onnx_type, [1, 1, 2, 2], [1.0, 1.0, 1.0, 1.0])
        ini_b = helper.make_tensor("bias", onnx_type, [1], [0.17])
        conv_tranpose_node = onnx.helper.make_node(
            "ConvTranspose",
            ["input", "weight", "bias"],
            ["output"],
            kernel_shape=[2, 2],
            output_padding=[0, 0],
            pads=[0, 0, 0, 0],
            strides=[1, 1],
            dilations=[1, 1],
            group=1,
        )
        graph = helper.make_graph(
            [conv_tranpose_node],
            "conv_transpose_test",
            [input_tensor],
            [output_tensor],
            initializer=[ini_w, ini_b],
        )
        model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", opset)])
        model.ir_version = ir_version  # use stable onnx ir version

        onnx.save(model, output_model_path)

    def static_quant_test_qdq(
        self,
        model_fp32_path,
        data_reader,
        activation_type,
        weight_type,
        extra_options=None,
    ):
        """
        Quantizes an FP32 model and checks various properties about the generated quantized model.

        :param model_fp32_path: The path to the FP32 ONNX model to quantize and test.
        :param data_reader: Data reader that generates input data.
        :param activation_type: The quantized type for activations. One of QuantType.QUInt8 or QuantType.QInt8.
        :param weight_type: The quantized type for weights. One of QuantType.QUInt8 or QuantType.QInt8.
        :param extra_options: Dictionary of extra options for the quantize_static() function.
        """

        activation_proto_qtype = TensorProto.UINT8 if activation_type == QuantType.QUInt8 else TensorProto.INT8
        activation_type_str = "u8" if (activation_type == QuantType.QUInt8) else "s8"
        weight_type_str = "u8" if (weight_type == QuantType.QUInt8) else "s8"
        model_int8_path = f"conv_transpose_fp32.quant_dqd_{activation_type_str}{weight_type_str}.onnx"

        data_reader.rewind()
        quantize_static(
            model_fp32_path,
            model_int8_path,
            data_reader,
            quant_format=QuantFormat.QDQ,
            activation_type=activation_type,
            weight_type=weight_type,
            extra_options=extra_options,
        )

        # Check node counts in quantized model.
        quant_node_counts = {"ConvTranspose": 1, "QuantizeLinear": 2, "DequantizeLinear": 4}
        check_op_type_count(self, model_int8_path, **quant_node_counts)

        # Check input/output types for QuantizeLinear nodes.
        qnode_io_qtypes = {
            "QuantizeLinear": [
                ["i", 2, activation_proto_qtype],
                ["o", 0, activation_proto_qtype],
            ]
        }
        check_qtype_by_node_type(self, model_int8_path, qnode_io_qtypes)

        # Check model correctness.
        data_reader.rewind()
        check_model_correctness(self, model_fp32_path, model_int8_path, data_reader.get_next())

    def quantize_conv_transpose_u8u8(self, onnx_type, opset, ir_version):
        """
        Unit test that quantizes (uint8) an ONNX model containing an ConvTranspose operator.
        """

        np.random.seed(1)
        model_fp32_path = "conv_transpose_fp32.onnx"
        self.construct_model(model_fp32_path, onnx_type, opset, ir_version)
        dtype = onnx.helper.tensor_dtype_to_np_dtype(onnx_type)
        data_reader = self.input_feeds(1, {"input": [1, 1, 7, 7]}, dtype)

        self.static_quant_test_qdq(
            model_fp32_path,
            data_reader,
            activation_type=QuantType.QUInt8,
            weight_type=QuantType.QUInt8,
        )

    def test_quantize_conv_transpose_u8u8(self):
        self.quantize_conv_transpose_u8u8(TensorProto.FLOAT, 13, 7)

    def test_quantize_conv_transpose_u8u8_fp16(self):
        self.quantize_conv_transpose_u8u8(TensorProto.FLOAT16, 19, 9)

    def quantize_conv_transpose_s8s8(self, onnx_type, opset, ir_version):
        """
        Unit test that quantizes (int8) an ONNX model containing an ConvTranspose operator.
        """

        np.random.seed(1)
        model_fp32_path = "conv_transpose_fp32.onnx"
        self.construct_model(model_fp32_path, onnx_type, opset, ir_version)
        dtype = onnx.helper.tensor_dtype_to_np_dtype(onnx_type)
        data_reader = self.input_feeds(1, {"input": [1, 1, 7, 7]}, dtype)

        self.static_quant_test_qdq(
            model_fp32_path,
            data_reader,
            activation_type=QuantType.QInt8,
            weight_type=QuantType.QInt8,
            extra_options={"ActivationSymmetric": True},
        )

    def test_quantize_conv_transpose_s8s8(self):
        self.quantize_conv_transpose_s8s8(TensorProto.FLOAT, 13, 7)

    def test_quantize_conv_transpose_s8s8_fp16(self):
        self.quantize_conv_transpose_s8s8(TensorProto.FLOAT16, 19, 9)


if __name__ == "__main__":
    unittest.main()
